;;;                 -*- Mode: Lisp; Package: BOOMS-APP -*-                  ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         abstract.lsp                                              ;;;
;;; Description:  BOOMS iconic abstractions                                 ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Date:         27-12-96                                                  ;;;
;;; Package:      BOOMS-APP                                                 ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  A B S T R A C T  file                                                  ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :booms-app)


;;;-------------------------------------------------------------------------;;;
;;;  I C O N I C - A B S T R A C T  classes                                 ;;;
;;;-------------------------------------------------------------------------;;;

;;; This is the iconic abstract class.
(defclass ICONIC-ABSTRACT (abstraction iconic-node) ())

(set-booms-name iconic-abstract 'abstract)

;;; This will be the class of abstractions with no values.
(defclass ICONIC-ABSTRACT-LEAF (abstraction iconic-leaf) ())

(set-booms-name iconic-abstract-leaf 'abstract-leaf)

;;; Icon for unnamed abstractions.
(defmethod ICON-ICON ((abstraction abstraction)) (booms-icon 25))

;;; Setting the arguments is fine (during abstraction it'll be set to nil).
(defmethod (setf ARGUMENTS) (newval (l iconic-abstract-leaf))
  t)

;;; Class changes of abstractions will happen - set the document modity flag.

(defmethod CHANGE-CLASS :after ((icon iconic-abstract) newclass
                                &rest initargs)
  (set-modified (icon-doc icon)))

(defmethod CHANGE-CLASS :after ((icon iconic-abstract-leaf) newclass
                                &rest initargs)
  (set-modified (icon-doc icon)))

;;; Describing an iconic-abstract-leaf with a string.
(defmethod ICON-DESCRIBE ((icon iconic-abstract-leaf) &optional socket)
  (format nil "~:(~a~) ~:[not-evaluated~*~;= ~a~]"
          (icon-describe-name icon)
          (evaluated-p icon) (and (evaluated-p icon) (evaluate icon))))

;;; These are used for allowing customization of abstraction classes.  The
;;; default returns the usual icon.  The first method makes passing t same as
;;; passing *current-type*.

(defmethod ICONIC-ABSTRACT-CLASS-NAME ((doctype (eql t)))
  (iconic-abstract-class-name *current-type*))

(defmethod ICONIC-ABSTRACT-CLASS-NAME ((doctype symbol))
  'iconic-abstract)

(defmethod ICONIC-ABSTRACT-LEAF-CLASS-NAME ((doctype symbol))
  'iconic-abstract-leaf)

;;; Create an abstraction - either an iconic-abstract or iconic-abstract-leaf.
(defun MAKE-ICON-ABSTRACT (node-list same-node-same-arg &rest initargs)
  (make-abstraction
   node-list
   same-node-same-arg
   (apply #'make-icon
          (if (have-abstract-args-p node-list)
            (iconic-abstract-class-name *current-type*)
            (iconic-abstract-leaf-class-name *current-type*))
          :centered t
          initargs)))


;;;-------------------------------------------------------------------------;;;
;;;  N A M E D - A B S T R A C T  class                                     ;;;
;;;-------------------------------------------------------------------------;;;

;;; Base class for named abstractions.
(defclass NAMED-ABSTRACTION ()
  ((abs-name :initarg :abs-name :reader abs-name)))

(defmethod ICON-ICON ((abstraction named-abstraction)) (booms-icon 26))

;;; Setting the abstraction name should redraw the icon, and modify the
;;; document.
(defmethod (setf ABS-NAME) ((newval string) (icon named-abstraction))
  (setf (slot-value icon 'abs-name) newval)
  (icon-reshow icon)
  (set-modified (icon-doc icon)))

;;; Making a copy - copy the name slot.
(defmethod MAKE-COPY ((icon named-abstraction))
  (let ((result (call-next-method)))
    (setf (abs-name result) (abs-name icon))
    result))

;;; Save the name slot.
(defmethod LOAD-SLOTS append ((x named-abstraction))
  '(abs-name))

;;; A predicate for named-abstractions.
(defmacro NAMED-ABSTRACT-P (obj)
  `(typep ,obj 'named-abstraction))

;;; The 'class name' of named abstractions is the name slot.
(defmethod ICON-DESCRIBE-NAME ((icon named-abstraction))
  (format nil "~s" (abs-name icon)))

;;; Drawing an icon for general named-abstraction - move the icon upwards, and
;;; draw the abstraction name.
(defmethod ICON-DRAW-ICON ((icon named-abstraction))
  (let ((pos    #.(make-position 0 0))
        (box    #.(make-box 0 0 0 0))
        (str    (abs-name icon))
        (doc    (icon-doc icon)))
    ;; Draw the original icon higher.
    (nmake-position pos
                    +icon-margin+ #.(+ +icon-margin+ (round +socket-size+ 2)))
    (draw-icon (icon-doc icon)
               (icon-icon icon)
               (nposition+ pos (nnbox-top-left (icon-box icon))))
    (nmake-position pos
                    +abstract-text-margin+ (- (box-height (icon-box icon)) 13))
    (nposition+ pos (nnbox-top-left (icon-box icon)))
    ;; Let some margin space at the bottom of leaves.
    (when (leafp icon)
      (decf (position-y pos) +abstract-text-margin+))
    (nmake-box-relative-from-corner
     box pos #.(- +icon-size+ (* +abstract-text-margin+ 2)) 13)
    (with-font (doc +abstract-text-font+)
      (with-foreground-color (doc +abstract-text-bg+)
        (fill-box doc box))
      (with-foreground-color (doc +abstract-text-fg+)
        (with-background-color (doc +abstract-text-bg+)
          (draw-string-in-box
           doc str 0 (length str) box
           (if (> (stream-string-width doc str) (box-width box)) :left :center)
           :center nil))))))

;;; The named abstractions.

(defclass ICONIC-NAMED-ABSTRACT (named-abstraction iconic-abstract)
  ())

(defclass ICONIC-NAMED-ABSTRACT-LEAF (named-abstraction iconic-abstract-leaf)
  ())

;;; Return the correct named abstraction class for an unnamed abstraction,
;;; based also on some doctype.  The first method makes passing t same as
;;; passing *current-type*.

(defmethod ICON-NAMED-ABSTRACTION-CLASS ((doctype (eql t))
                                         (icon iconic-abstract))
  (icon-named-abstraction-class *current-type* icon))

(defmethod ICON-NAMED-ABSTRACTION-CLASS ((doctype symbol)
                                         (icon iconic-abstract))
  'iconic-named-abstract)

(defmethod ICON-NAMED-ABSTRACTION-CLASS ((doctype symbol)
                                         (icon iconic-abstract-leaf))
  'iconic-named-abstract-leaf)

;;; Return the correct unnamed abstraction class for a named abstraction, based
;;; also on some doctype.  The first method makes passing t same as passing
;;; *current-type*.

(defmethod ICON-UNNAMED-ABSTRACTION-CLASS ((doctype (eql t))
                                           (icon iconic-named-abstract))
  (icon-unnamed-abstraction-class *current-type* icon))

(defmethod ICON-UNNAMED-ABSTRACTION-CLASS ((doctype symbol)
                                           (icon iconic-named-abstract))
  'iconic-abstract)

(defmethod ICON-UNNAMED-ABSTRACTION-CLASS ((doctype symbol)
                                           (icon iconic-named-abstract-leaf))
  'iconic-abstract-leaf)


;;;-------------------------------------------------------------------------;;;
;;;  N A M E D - C O P Y  class                                             ;;;
;;;-------------------------------------------------------------------------;;;

;;; The named copy class is a named-abstraction subclass that simply contains
;;; a list of icons to paste.  It is not an icon - it is only a place for
;;; storing the icon list and associate it with a name.
(defclass NAMED-COPY (named-abstraction)
  ((icon-list :initarg :icon-list :initform nil :accessor icon-list)))

;;; Setting the named copy name should not redraw an icon.
(defmethod (setf ABS-NAME) ((newval string) (icon named-copy))
  (setf (slot-value icon 'abs-name) newval))

;;; Create a named-copy - no icons involved.
(defun MAKE-NAMED-COPY (node-list name &optional dont-copy)
  (let ((icons (if dont-copy node-list (copy-nodes node-list))))
    (setf (icons-center-position icons) #.(make-position 0 0))
    (make-instance 'named-copy :abs-name name :icon-list icons)))


;;;-------------------------------------------------------------------------;;;
;;;  N A M E D - A B S T R A C T I O N  operations                          ;;;
;;;-------------------------------------------------------------------------;;;

;;; A plist of doc types and their created abstractions.
(defvar *BOOMS-ABSTRACTIONS* nil)

;;; A plist of doc-types and their protected abstractions - used for not
;;; deleting named copies that are used as classes.
(defvar *PROTECTED-ABSTRACTIONS* nil)

;;; Does an abstraction name already exists (in any type)?
(defun ABSTRACT-NAME-EXISTS-P (name doc-type)
  (member name (getf *booms-abstractions* doc-type)
          :test #'equalp :key #'first))

;;; Update the abstractions-list if needed, do nothing if not a named
;;; abstraction.

(defmethod NEW-NAMED-ABSTRACTION ((icon icon) &key name doc-type (make-copy t))
  nil)

(defmethod NEW-NAMED-ABSTRACTION ((icon named-abstraction)
                                  &key name doc-type (make-copy t)
                                       (path-place nil)
                                       (registering nil))
  (when name
    (let ((doc (icon-doc icon)))
      (setf (abs-name icon) name)
      (do-redraws doc)))
  (let ((doc-type (or doc-type (doc-type (icon-doc icon))))
        (name (abs-name icon)))
    (unless (abstract-name-exists-p name doc-type)
      (push (list name (if make-copy (make-copy icon) icon))
            (getf *booms-abstractions* doc-type))
      (when registering
        (push name (getf *protected-abstractions* doc-type)))
      (if registering
        (apply #'add-node-class name path-place)
        (add-node-class name doc-type (or path-place :user-defined))))))

;;; Rename an abstraction - update the menu and the outline.
(defmethod RENAME-NAMED-ABSTRACTION ((icon named-abstraction) (newname string))
  (let ((oldname (abs-name icon))
        (doc-type (doc-type (icon-doc icon))))
    (dolist (doc (get-all-documents))
      (when (eql (doc-type doc) doc-type)
        (dolist (icon (icon-list doc))
          (when (and (named-abstract-p icon)
                     (equalp (abs-name icon) oldname))
            (setf (abs-name icon) newname)))
        (do-redraws doc)))
    (rename-node-class oldname newname doc-type :user-defined)
    (let ((abs (find oldname (getf *booms-abstractions* doc-type)
                     :key #'first :test #'equalp)))
      (setf (first abs) newname)
      (setf (abs-name (second abs)) newname))
    (when (equalp *current-node-class* oldname)
      (set-current-node-class newname))))

;;; When an abstraction will be added - it'll use the name string as the class
;;; to generate.  This method makes this call create the named abstraction
;;; object or the copy of a named-copy object.
(defmethod MAKE-ICON ((class string)
                      &rest initargs
                      &key (document nil) (pos nil) (centered nil)
                      &allow-other-keys)
  (let ((abstract (second
                   (find class (getf *booms-abstractions* *current-type*)
                         :test #'equalp :key #'first))))
    (unless abstract
      (error "Abstraction ~s not found in abstractions for ~s"
             class *current-type*))
    (let (result)
      (cond
        ((typep abstract 'named-copy)
         (setf result (copy-nodes (icon-list abstract)))
         (dolist (icon result)
           (setf (icon-doc icon) document))
         (when pos
           (if centered
             (dolist (icon result)
               (icon-move-relative icon pos))
             (dolist (icon result)
               ;; This should not happen...
               (nbox-move (icon-box icon) (nbox-top-left (icon-box icon)))
               (icon-move-relative icon pos)))))
        (t
         (setf result (make-copy abstract))
         (setf (icon-doc result) document)
         (when pos
           (if centered
             (icon-move result pos)
             (progn
               (nbox-move (icon-box result) (nbox-top-left (icon-box result)))
               (icon-move-relative result pos))))))
      result)))


;;;-------------------------------------------------------------------------;;;
;;;  C L I P B O A R D  as a named-copy                                     ;;;
;;;-------------------------------------------------------------------------;;;

;;; The clipboard class name.
(defconstant +CLIPBOARD-NAME+ "ClipBoard")

;;; This class is used to hold the clipboard of some document type, for making
;;; it like normal named-copy.
(defclass CLIPBOARD-NAMED-COPY (named-copy) ())

;;; This is called when copying to add the special clipboard entry.
(defun ADD-CLIPBOARD-NAMED-COPY (doc icons)
  (let ((doc-type (doc-type doc)))
    (unless (abstract-name-exists-p +clipboard-name+ doc-type)
      (let ((*registering-type* t)) ; Prevent selecting this type as current.
        (new-named-abstraction
         (make-instance 'clipboard-named-copy :abs-name +clipboard-name+)
         :doc-type doc-type :make-copy nil :path-place :user-copies)))
    (push +clipboard-name+ (getf *protected-abstractions* doc-type))
    (let ((clipboard-icon
           (second (find +clipboard-name+ (getf *booms-abstractions* doc-type)
                         :test #'equalp :key #'first))))
      (setf (icon-list clipboard-icon) icons))))


;;;-------------------------------------------------------------------------;;;
;;;  D O C U M E N T  interface functions                                   ;;;
;;;-------------------------------------------------------------------------;;;

;;; Build a group icon.
(defmethod DO-ABSTRACT ((doc document))
  (with-lock *prevent-actions*
    (when (some #'selected (icon-list doc))
      (let* ((selected-icons (selected-icon-list doc))
             abs)
        (when (or (> (length selected-icons) 1)
                  (my-yes-or-no-p
                   "Are you sure you want to abstract a single icon?~@
                    (you might want to use a named copy)"))
          (unhilight-current-icon doc)
          (dolist (icon (icon-list doc))
            (when (selected icon)
              (icon-lines-hide icon)))
          (setf abs (make-icon-abstract
                     (selected-icon-list doc)
                     (same-node-same-arg-on doc)
                     :document doc
                     :pos (icons-center-position selected-icons)))
          (if abs
            (progn
              (dolist (icon selected-icons)
                (when (selected icon)
                  (icon-hide icon)))
              (setf (icon-list doc) (delete-if #'selected (icon-list doc)))
              (push abs (icon-list doc))
              (setf (icons-center-position selected-icons)
                    #.(make-position 0 0))
              (when (grid-on doc) (icon-to-grid abs))
              (icon-show abs)
              (fix-doc-size doc)
              (select-one-icon doc abs)
              (do-redraws doc)
              (hilight-current-icon doc t)
              (say "Abstraction done."))
            (progn
              (dolist (icon (icon-list doc))
                (when (selected icon)
                  (icon-lines-show icon)))
              (hilight-current-icon doc t)
              (say! "Abstraction failed!"))))))))

;;; Explode a group icon.
(defmethod DO-EXPLODE ((doc document))
  (with-lock *prevent-actions*
    (when (some #'(lambda (x) (and (selected x) (abstractionp x)))
                (icon-list doc))
      (unhilight-current-icon doc)
      (dolist (icon
               (remove-if-not
                #'(lambda (x) (and (selected x) (abstractionp x)))
                (icon-list doc)))
        (let ((internal-nodes nil))
          (icon-lines-hide icon)
          (mapc #'icon-lines-hide (uplinks icon))
          (icon-hide icon)
          (setf internal-nodes (explode-abstraction icon))
          (setf (icon-list doc)
                (append internal-nodes (delete icon (icon-list doc))))
          (setf (icons-center-position internal-nodes) (icon-center-pos icon))
          (dolist (icon internal-nodes)
            (setf (icon-doc icon) doc)
            (reset-redraw-info icon)
            (icon-lines-hide icon)
            (icon-select icon)
            (icon-show icon)
            (when (grid-on doc) (icon-to-grid icon)))))
      (fix-doc-size doc)
      (do-redraws doc)
      (hilight-current-icon doc t)
      (say "Explode done."))))

;;; Name an abstraction.
(defmethod DO-NAME-ABSTRACTION ((doc document))
  (with-lock *prevent-actions*
    (if (and (= (count-if #'selected (icon-list doc)) 1)
             (abstractionp (find-if #'selected (icon-list doc))))
      (let ((icon (find-if #'selected (icon-list doc))))
        (if (named-abstract-p icon)
          (when (my-yes-or-no-p
                 "Are you sure you want to rename this abstraction?~@
                  (saved files will not be modified)")
            (let* ((oldname (abs-name icon))
                   (newname (dialog-get-string "Select a new name" oldname)))
              (cond
                ((or (not newname)
                     (equalp newname "") (equalp newname oldname))
                 (hilight-current-icon doc t)
                 (say "Rename canceled."))
                ((abstract-name-exists-p newname (doc-type doc))
                 (hilight-current-icon doc t)
                 (say! "Abstraction name already exists, renaming canceled."))
                (t
                 (rename-named-abstraction icon newname)
                 (hilight-current-icon doc t)
                 (say "Abstraction ~s renamed to ~s." oldname newname)))))
          (let ((name
                 (dialog-get-string "Select a name for this abstraction" "")))
            (cond
              ((abstract-name-exists-p name (doc-type doc))
               (hilight-current-icon doc t)
               (say! "Abstraction name already exists, naming canceled."))
              ((and name (not (equalp name "")))
               (change-class
                icon (icon-named-abstraction-class *current-type* icon))
               (new-named-abstraction icon :name name)
               (hilight-current-icon doc t)
               (say "Named-Abstraction ~s created." name))))))
      (when (some #'selected (icon-list doc))
        (let ((name
               (progn
                 (say "An abstraction will be created first")
                 (dialog-get-string "Select a name for this abstraction" ""))))
          (cond
            ((abstract-name-exists-p name (doc-type doc))
             (hilight-current-icon doc t)
             (say! "Abstraction name already exists, naming canceled."))
            ((and name (not (equalp name "")))
             (let ((*prevent-actions* nil))
               (do-abstract doc))
             (if (and (= (count-if #'selected (icon-list doc)) 1)
                      (abstractionp (find-if #'selected (icon-list doc))))
               (let ((icon (find-if #'selected (icon-list doc))))
                 (change-class
                  icon (icon-named-abstraction-class *current-type* icon))
                 (new-named-abstraction icon :name name)
                 (hilight-current-icon doc t)
                 (say "Named-Abstraction ~s created." name))
               (progn
                 (hilight-current-icon doc t)
                 (say "Abstraction canceled."))))))))))

;;; Unnamed the selected icons.
(defmethod DO-UNNAME-ABSTRACTION ((doc document))
  (with-lock *prevent-actions*
    (let ((changep nil))
      (dolist (icon (selected-icon-list doc))
        (when (named-abstract-p icon)
          (change-class
           icon (icon-unnamed-abstraction-class *current-type* icon))
          (icon-reshow icon)
          (setf changep t)))
      (hilight-current-icon doc t)
      (if changep
        (progn
          (do-redraws doc)
          (say "Unnaming done."))
        (say "No nodes to unname")))))

;;; Create a named copy.
(defmethod DO-CREATE-NAMED-COPY ((doc document))
  (with-lock *prevent-actions*
    (let ((icons (selected-icon-list doc)))
      (unless (null icons)
        (let ((name
               (dialog-get-string "Select a name for this named copy" "")))
          (hilight-current-icon doc t)
          (if (abstract-name-exists-p name (doc-type doc))
            (say! "Abstraction name already exists, naming canceled.")
            (progn
              (new-named-abstraction (make-named-copy icons name)
                                     :doc-type (doc-type doc) :make-copy nil
                                     :path-place :user-copies)
              (say "Named copy created."))))))))

;;; Remove the current user class - an abstraction or a named-copy.
(defmethod DO-REMOVE-CURRENT-USER-NODE ((doc document))
  (with-lock *prevent-actions*
    (let* ((doc-type (doc-type doc)))
      (cond
        ((or (member *current-node-class*
                     (getf *protected-abstractions* doc-type)
                     :test #'equalp)
             (not (abstract-name-exists-p *current-node-class* doc-type)))
         (hilight-current-icon doc t)
         (say! "Selected class is not a user class."))
        ((typep (second (find *current-node-class*
                              (getf *booms-abstractions* doc-type)
                              :test #'equalp :key #'first))
                'named-copy)
         (delete-node-class *current-node-class* doc-type :user-copies)
         (setf (getf *booms-abstractions* doc-type)
               (delete *current-node-class*
                       (getf *booms-abstractions* doc-type)
                       :test #'equalp :key #'first))
         (set-first-available-node-class-for doc-type)
         (refresh-class-palette)
         (hilight-current-icon doc t)
         (say "Named copies removed."))
        (t
         (let ((abs-used-p
                (block search-docs
                  (dolist (doc (get-all-documents))
                    (when (eql (doc-type doc) doc-type)
                      (do* ((icons (icon-list doc))
                            (icon (pop icons) (pop icons)))
                          ((null icon))
                        (when (and (named-abstract-p icon)
                                   (equalp *current-node-class*
                                           (abs-name icon)))
                          (return-from search-docs t))
                        (when (abstractionp icon)
                          (setf icons (append (internal-nodes icon) icons))))))
                  nil)))
           (if (or (not abs-used-p)
                   (my-yes-or-no-p "The user-defined abstraction ~s is used,~@
                                    Are you sure you want to delete it? ~
                                    (unnaming all instances)"
                                   *current-node-class*))
             (progn
               (when abs-used-p
                 (dolist (doc (get-all-documents))
                   (when (eql (doc-type doc) doc-type)
                     (do* ((icons (icon-list doc))
                           (icon (pop icons) (pop icons)))
                         ((null icon))
                       (when (and (named-abstract-p icon)
                                  (equalp *current-node-class*
                                          (abs-name icon)))
                         (change-class
                          icon
                          (icon-unnamed-abstraction-class *current-type* icon))
                         (icon-reshow icon))
                       (when (abstractionp icon)
                         (setf icons (append (internal-nodes icon) icons))))
                     (do-redraws doc))))
               (delete-node-class *current-node-class* doc-type :user-defined)
               (setf (getf *booms-abstractions* doc-type)
                     (delete *current-node-class*
                             (getf *booms-abstractions* doc-type)
                             :test #'equalp :key #'first))
               (set-first-available-node-class-for doc-type)
               (refresh-class-palette)
               (hilight-current-icon doc t)
               (say "Abstraction removed."))
             (progn
               (hilight-current-icon doc t)
               (say "Deletion canceled.")))))))))

;;; Remove all unused abstractions.
(defmethod DO-REMOVE-UNUSED-ABSTRACTIONS ((doc document))
  (with-lock *prevent-actions*
    (let ((doc-type (doc-type doc))
          (abs-names nil))
      (dolist (doc (get-all-documents))
        (when (eql (doc-type doc) doc-type)
          (do* ((icons (icon-list doc))
                (icon (pop icons) (pop icons)))
              ((null icon))
            (when (named-abstract-p icon)
              (pushnew (abs-name icon) abs-names))
            (when (abstractionp icon)
              (setf icons (append (internal-nodes icon) icons))))))
      (setf abs-names
            (set-difference
             (mapcar #'first
                     (remove-if #'(lambda (x) (typep (second x) 'named-copy))
                                (getf *booms-abstractions* doc-type)))
             abs-names))
      (cond
        ((null abs-names)
         (hilight-current-icon doc t)
         (say "No unused abstractions found."))
        ((my-yes-or-no-p
          "Are you sure you want to remove these named abstracts:~%~{  ~a~^,~}"
          abs-names)
         (dolist (name abs-names)
           (delete-node-class name doc-type :user-defined)
           (setf (getf *booms-abstractions* doc-type)
                 (delete name (getf *booms-abstractions* doc-type)
                         :test #'equalp :key #'first)))
         (when (member *current-node-class* abs-names)
           (set-first-available-node-class-for doc-type))
         (refresh-class-palette)
         (hilight-current-icon doc t)
         (say "Named abstractions removed."))
        (t
         (hilight-current-icon doc t)
         (say "Deletion canceled."))))))

;;; Remove all named-copy entries, except the ClipBoard entry.
(defmethod DO-REMOVE-NAMED-COPIES ((doc document))
  (with-lock *prevent-actions*
    (let* ((doc-type (doc-type doc))
           (copy-names
            (nset-difference
             (mapcar #'first
                     (remove-if-not
                      #'(lambda (x) (typep (second x) 'named-copy))
                      (getf *booms-abstractions* doc-type)))
             (getf *protected-abstractions* doc-type)
             :test #'equalp)))
      (cond
        ((null copy-names)
         (hilight-current-icon doc t)
         (say "No named copies found."))
        ((my-yes-or-no-p
          "Are you sure you want to remove these named copies:~%~{  ~a~^,~}"
          copy-names)
         (dolist (name copy-names)
           (delete-node-class name doc-type :user-copies)
           (setf (getf *booms-abstractions* doc-type)
                 (delete name (getf *booms-abstractions* doc-type)
                         :test #'equalp :key #'first)))
         (when (member *current-node-class* copy-names)
           (set-first-available-node-class-for doc-type))
         (refresh-class-palette)
         (hilight-current-icon doc t)
         (say "Named copies removed."))
        (t
         (hilight-current-icon doc t)
         (say "Deletion canceled."))))))


;;;-------------------------------------------------------------------------;;;
;;;  N A M E D - C O P I E S  as initial node classes                       ;;;
;;;-------------------------------------------------------------------------;;;

;;; This function is used to define a built-in named-copy, as a template value.
;;; It should be called during registration.
(defun ADD-NODE-COPY-CLASS (copy-name icon-class icon-init-args &rest path)
  (when *registering-type*
    (new-named-abstraction
     (make-named-copy
      (list
       (apply #'make-icon icon-class :pos (make-position 0 0) icon-init-args))
      copy-name t)
     :doc-type    *registering-type*
     :make-copy   nil
     :path-place  path
     :registering t)))


;;;-------------------------------------------------------------------------;;;
;;;  A B S T R A C T I O N S  icon menu                                     ;;;
;;;-------------------------------------------------------------------------;;;

;;; The menu variable.
(defvar *ABSTRACTION-ICON-MENU* nil)

;;; Return this menu for iconic abstractions.
(defmethod ICON-MENU ((icon iconic-abstract))
  *abstraction-icon-menu*)

;;; Return this menu for iconic abstraction leaves also.
(defmethod ICON-MENU ((icon iconic-abstract-leaf))
  *abstraction-icon-menu*)

;;; Create the menu.
(defun MAKE-ABSTRACTION-ICON-MENU ()
  (let ((menu (make-default-icon-menu)))
    (add-icon-menu-item menu
                        #'(lambda (icon)
                            (let ((*prevent-actions* nil))
                              (do-explode (icon-doc icon))))
                        "E~xplode" "Explode this abstraction")
    (add-icon-menu-item menu
                        #'(lambda (icon)
                            (let ((*prevent-actions* nil))
                              (do-name-abstraction (icon-doc icon))))
                        "~Name" "Name/Rename this abstraction")
    (add-icon-menu-item menu
                        #'(lambda (icon)
                            (let ((*prevent-actions* nil))
                              (do-unname-abstraction (icon-doc icon))))
                        "~Unname" "Uname this abstraction (if its named)")
    menu))


;;;-------------------------------------------------------------------------;;;
;;;  I N I T I A L I Z A T I O N  function                                  ;;;
;;;-------------------------------------------------------------------------;;;

;;; Initialize the abstractions system.
(defun INITIALIZE-ABSTRACTIONS  ()
  (setf *booms-abstractions* nil)
  (setf *abstraction-icon-menu* (make-abstraction-icon-menu)))


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :ABSTRACT)

