;;;                 -*- Mode: Lisp; Package: BOOMS-APP -*-                  ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         boomsapp.lsp                                              ;;;
;;; Description:  BOOMS windows application                                 ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Idea by:      Omri Weisman - omri@cs.bgu.ac.il                          ;;;
;;;               Cohavit Taboch  - taboch@cs.bgu.ac.il                     ;;;
;;; Date:         29-12-96                                                  ;;;
;;; Package:      BOOMS-APP                                                 ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  B O O M S A P P  file                                                  ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :booms-app)


;;;-------------------------------------------------------------------------;;;
;;;  G L O B A L  variables                                                 ;;;
;;;-------------------------------------------------------------------------;;;

;;; This is the Booms application - the window object.
(defvar *BOOMS-APP*                   nil)

;;; The menu, toolbar and status-bar.
(defvar *BOOMS-MENU*                  nil)
(defvar *BOOMS-TOOLBAR*               nil)
(defvar *BOOMS-STATUS-BAR*            nil)

;;; The shown flags for the toolbar and status-bar.
(defvar *TOOLBAR-SHOWN*               nil)
(defvar *STATUS-BAR-SHOWN*            nil)

;;; List of pop-up menus to destroy when exiting BOOMS.
(defvar *MISC-MENUS*                  nil)

;;; The right-button menu for the application background.
(defvar *APP-RIGHT-MENU*              nil)

;;; The right-button menu for the document background.
(defvar *DOC-RIGHT-MENU*              nil)

;;; The BOOMS clipboard - a plist of type and contents.
(defvar *BOOMS-CLIPBOARD*             nil)

;;; List of node classes used.
(defvar *NODE-CLASSES*                nil)

;;; Current node clas (selected from menu).
(defvar *CURRENT-NODE-CLASS*          nil)

;;; The icons menu & outline-box to add node classes into.
(defvar *BOOMS-ICON-CLASS-MENU*       nil)
(defvar *BOOMS-ICON-CLASS-OUTLINE*    nil)

;;; The list of menu items in the menu.
(defvar *BOOMS-ICON-CLASS-MENU-ITEMS* nil)

;;; The doc-types declared.
(defvar *BOOMS-DOC-TYPES*             nil)

;;; The doc type registering now.
(defvar *REGISTERING-TYPE*            nil)

;;; The current active document's type.
(defvar *CURRENT-TYPE*                nil)

;;; This is to remember the last class for a document type.
(defvar *LAST-CLASS-FOR-TYPES*        nil)

;;; Untitled window name counter.
(defvar *UNTITLED-COUNTER*            nil)

;;; Document-dependent menu / dialog items (unavaile when none).
(defvar *DOC-DEPENDENT-ITEMS*         nil)

;;; The list of logical document numbers currently open.
(defvar *BOOMS-DOC-N-LIST*            nil)

;;; This is the currently grabbed style.
(defvar *GRABBED-STYLE*               nil)

;;; This will be the last "say" string, or nil if "unsay".
(defvar *LAST-SAID*                   nil)

;;; The second status bar text box (for the say2 function).
(defvar *SAY2-TEXT*                   nil)

;;; This will be t whenever set-value for an item is called programmatically.
(defvar *CALLED-PROGRAMMATICALLY*     nil)

;;; When true - prevent the user-close method from closing a document.  Used to
;;; first ask all document-save questions before quitting.
(defvar *DONT-CLOSE*                  nil)

;;; When true - prevent the user-close method from asking to save a document.
(defvar *DONT-ASK*                    nil)

;;; When we happen to terminate without our throw (in case of a normal image)
;;; we don't want a throw without a catch...
(defvar *DONT-THROW-DONE*             t)

;;; When we are loading a file this will be bound to t, this is for the icon
;;; creation to keep quiet and not report errors.
(defvar *LOADING-NOW*                 nil)

;;; This variable is needed for not displaying a message on a mouse-in event
;;; after activating some menu.
(defvar *PREVENT-NEXT-READY-MESSAGE*  nil)

;;; The mode item variables, each a cons of the menu-item and the button.
(defvar *SAME-NODE-SAME-ARG-ITEMS*    nil)
(defvar *DIALOG-CREATE-ITEMS*         nil)
(defvar *GRID-ITEMS*                  nil)

;;; The windows menu to add window items into.
(defvar *BOOMS-WINDOW-MENU*           nil)

;;; The toolbar next button's offset.
(defvar *TOOLBAR-BUTTON-OFFSET*       nil)

;;; The next document winow location.
(defvar *WIN-TOP*                     nil)
(defvar *WIN-LEFT*                    nil)


;;;-------------------------------------------------------------------------;;;
;;;  M I S C  definitions                                                   ;;;
;;;-------------------------------------------------------------------------;;;

;;; Get the active BOOMS application pane.
(defun GET-BOOMS-PANE (&optional (*booms-app* *booms-app*))
  (frame-child *booms-app*))

;;; Return all existing documents.
(defun GET-ALL-DOCUMENTS (&optional (*booms-app* *booms-app*))
  (mapcar #'doc (windows (get-booms-pane))))

;;; Get the active document.
(defun ACTIVE-DOCUMENT (&optional (*booms-app* *booms-app*))
  (let ((active-frame (front-window (get-booms-pane *booms-app*))))
    (and active-frame (doc active-frame))))

;;; Select the next window.
(defun SELECT-NEXT-DOC ()
  (let ((docs (get-all-documents)))
    (when docs
      (select-doc (first (last docs))))))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N S  handling                                                    ;;;
;;;-------------------------------------------------------------------------;;;

;;; A hash-table with loaded icons.
(defvar *BOOMS-ICONS* nil)

;;; Function to extract icons from the "icons.dll" icons file.
(defun BOOMS-ICON (index)
  (or (gethash index *booms-icons*)
      (setf (gethash index *booms-icons*)
            (extract-icon-from-file "icons.dll" index))))


;;;-------------------------------------------------------------------------;;;
;;;  D O C U M E N T - T Y P E  handling                                    ;;;
;;;-------------------------------------------------------------------------;;;

;;; Handling registration of document type symbols.
(defun REGISTER-TYPE (type)
  (setf *registering-type* type)
  (pushlast type *booms-doc-types*)
  (add-shared-classes))

;;; Switch types.
(defun SET-CURRENT-TYPE (doc)
  (let ((newtype (doc-type doc))
        (outline *booms-icon-class-outline*))
    (when (and newtype (not (eq *current-type* newtype)))
      (outline-change-value-availability outline newtype t)
      (outline-open-all-value            outline newtype)
      (outline-change-value-availability outline *current-type* nil)
      (outline-close-value               outline *current-type*)
      (dolist (item (menu-items *booms-icon-class-menu*))
        (set-menu-item-available-p item (eq (menu-item-title item) newtype)))
      (setf (getf *last-class-for-types* *current-type*) *current-node-class*)
      ;; This is done first for not getting back here in case that
      ;; set-current-node-class will cause the document to get the focus again.
      (setf *current-type* newtype)
      (if (getf *last-class-for-types* newtype)
        (set-current-node-class (getf *last-class-for-types* newtype) t)
        (set-first-available-node-class-for newtype t))
      (say2 newtype))))

;;; Disable type widgets.
(defun DISABLE-TYPE-WIDGETS ()
  (let ((outline *booms-icon-class-outline*))
    (setf (getf *last-class-for-types* *current-type*) *current-node-class*)
    (setf *current-type* nil)
    (dolist (type *booms-doc-types*)
      (outline-change-value-availability outline type nil)
      (outline-close-value outline type))
    (set-current-node-class nil))
  (say2 ""))


;;;-------------------------------------------------------------------------;;;
;;;  D O C - D E P E N D A N T  handling                                    ;;;
;;;-------------------------------------------------------------------------;;;
;;; Document-dependent items are unavailable when no document is open.

;;; Add a menu item to the document-dependent list.
(defun DOC-DEPENDENT (item)
  (push item *doc-dependent-items*)
  item)

;;; Setting something to be available / unavailable.
;;; Menu item - simply disable it.
(defmethod SET-ITEM-AVAILABLE-P ((mi menu-item) value)
  (set-menu-item-available-p mi value))
;;; Dialog-item - make the set-value-fn do nothing (will still respond to
;;; mouse-in events).
(defmethod SET-ITEM-AVAILABLE-P ((di dialog-item) value)
  (setf (dialog-item-set-value-fn di)
        (if value
          (getf (dialog-item-plist di) :saved-set-value-fn)
          (progn
            (setf (getf (dialog-item-plist di) :saved-set-value-fn)
                  (dialog-item-set-value-fn di))
            #'(lambda (&rest args) (say "No active document!") nil)))))

;;; Added the first document, make all doc-dependent items available.
(defun ENABLE-DOC-DEPENDENTS ()
  (dolist (menu-item *doc-dependent-items*)
    (set-item-available-p menu-item t)))

;;; Last document removed, make all mode items nil and doc-dependent items
;;; unavailable.
(defun DISABLE-DOC-DEPENDENTS ()
  (set-mode-items nil)
  (dolist (menu-item *doc-dependent-items*)
    (set-item-available-p menu-item nil)))


;;;-------------------------------------------------------------------------;;;
;;;  M O D E  menu items & buttons handling                                 ;;;
;;;-------------------------------------------------------------------------;;;
;;; This handles menu items and buttons which represent some kind of a mode
;;; that should be set/unset.  And item variable is a cons of a menu item and a
;;; dialog item.

;;; Set the mode items for the current document.
(defun SET-MODE-ITEMS (doc)
  (set-mode-items-value *same-node-same-arg-items*
                        (and doc (same-node-same-arg-on doc)))
  (set-mode-items-value *dialog-create-items* (and doc (dialog-create-on doc)))
  (set-mode-items-value *grid-items*          (and doc (grid-on doc))))

;;; Set mode item values without activating another value setting.
(defun SET-MODE-ITEMS-VALUE (mode-items value)
  (let ((*called-programmatically* t))
    (when (first mode-items)
      (set-menu-item-selected-p (first mode-items) value))
    (when (rest mode-items)
      (set-dialog-item-value    (rest mode-items) value))))


;;;-------------------------------------------------------------------------;;;
;;;  D O C - W I N D O W S  handling of creating / destroying               ;;;
;;;-------------------------------------------------------------------------;;;

;;; Handle new untitled window titles.
(defun GET-NEXT-UNTITLED-TITLE (type)
  (let ((typestr (capitalize-object type)))
    (format nil "~:(~a~)~d.bms"
            (subseq typestr 0 (min 5 (length typestr)))
            (incf *untitled-counter*))))

;;; Calculate exterior of new document frame.
(defun GET-NEW-EXTERIOR (parent)
  (let* ((parent-width  (window-exterior-width parent))
         (parent-height (window-exterior-height parent))
         (page-width    (floor (* parent-width 0.75)))
         (page-height   (floor (* parent-height 0.75))))
    (incf *win-left* (get-caption-height))
    (incf *win-top*  (get-caption-height))
    (when (> (+ *win-left* page-width)  parent-width)  (setf *win-left* 0))
    (when (> (+ *win-top*  page-height) parent-height) (setf *win-top*  0))
    (make-box *win-left* *win-top*
              (+ page-width *win-left*) (+ page-height *win-top*))))

;;; Return a logical number for a newly created document window.
(defun GET-NEW-DOCUMENT-NUMBER ()
  (do ((i 1 (1+ i)))
      ((not (member i *booms-doc-n-list*)) i)))

;;; Return the menu-item title for some document.
(defun DOC-MENU-TITLE (title l-number)
  (format nil "~~~36r ~a" l-number title))

;;; Return an event-synonym to select a document window.
(defun DOC-MENU-EVENT-SYNONYM (l-number)
  (if (< l-number 10)
    (list 'alt-key (elt "0123456789" l-number))
    nil))

;;; Added a document, handle menu items.
(defun DOCUMENT-ADDED (doc)
  (when (null *booms-doc-n-list*)
    (first-doc-added))
  (let ((new-n (get-new-document-number)))
    (push new-n *booms-doc-n-list*)
    (add-to-menu
     *booms-window-menu*
     (make-menu-item :title         (doc-menu-title (title doc) new-n)
                     :value         #'(lambda () (select-doc doc))
                     :event-synonym (doc-menu-event-synonym new-n)
                     :help-string   (format nil "Select window \"~a\""
                                            (title doc))
                     :plist         (list :doc doc :number new-n)))
    (document-got-focused doc)))

;;; Removed a document, handle menu items.
(defun DOCUMENT-REMOVED (doc)
  (let ((menu-item (find-if
                    #'(lambda (mi) (eq (getf (menu-item-plist mi) :doc) doc))
                    (menu-items *booms-window-menu*))))
    (setf *booms-doc-n-list*
          (delete (getf (menu-item-plist menu-item) :number)
                  *booms-doc-n-list*))
    (remove-from-menu *booms-window-menu* menu-item)
    (when (null *booms-doc-n-list*)
      (last-doc-removed))))

;;; Handle first and last document cases.

(defun FIRST-DOC-ADDED ()
  (enable-doc-dependents))

(defun LAST-DOC-REMOVED ()
  (disable-doc-dependents)
  (disable-type-widgets))

;;; Document has changed its title.
(defun DOCUMENT-TITLE-CHANGED (doc)
  (let ((menu-item (find-if #'(lambda (mi)
                                (eq (getf (menu-item-plist mi) :doc) doc))
                            (menu-items *booms-window-menu*))))
    (when menu-item
      (set-menu-item-title menu-item (title doc)))))

;;; Document has got the focus.
(defun DOCUMENT-GOT-FOCUSED (doc)
  (dolist (mi (menu-items *booms-window-menu*))
    (set-menu-item-selected-p mi (eq (getf (menu-item-plist mi) :doc) doc)))
  (set-mode-items doc)
  (set-current-type doc))


;;;-------------------------------------------------------------------------;;;
;;;  M E N U  creation                                                      ;;;
;;;-------------------------------------------------------------------------;;;

;;; Function to create menu items.
(defun MENU-ENTRY (command)
  (let ((item (make-menu-item :title         (get command :title)
                              :value         command
                              :event-synonym (get command :accelerator nil)
                              :help-string   (get command :help-string))))
    (if (get command :doc-dependent nil)
      (doc-dependent item)
      item)))

;;; Classes for my menu items to say their help string.

(defclass MY-MENU () ())

(defclass MY-MENU-BAR       (my-menu menu-bar)       ())

(defclass MY-PULL-DOWN-MENU (my-menu pull-down-menu) ())

;;; Create the menu-bar on the given window.
(defun MAKE-BOOMS-MENU (win)
  (let ((menu-bar (open-stream 'my-menu-bar win :io)))
    ;; File menu.
    (add-to-menu menu-bar
     (make-menu-item :title "~File"
      :help-string "File related operations"
      :value
      (open-menu (list (menu-entry 'file-new)
                       (menu-entry 'file-open)
                       (menu-entry 'file-close)
                       (menu-entry 'file-close-all)
                       (menu-entry 'file-save)
                       (menu-entry 'file-save-as)
                       (menu-entry 'file-save-all)
                       menu-separator
                       (menu-entry 'file-exit))
                 'my-pull-down-menu *screen*
                 :selection-function 'funcall-menu-item)))
    ;; Edit menu.
    (add-to-menu menu-bar
     (doc-dependent
      (make-menu-item :title "~Edit"
       :help-string "Icon editing operations"
       :value
       (open-menu (list (menu-entry 'edit-cut)
                        (menu-entry 'edit-copy)
                        (menu-entry 'edit-paste)
                        (menu-entry 'edit-delete)
                        (menu-entry 'edit-edit-icon)
                        menu-separator
                        (menu-entry 'edit-grab-style)
                        (menu-entry 'edit-apply-style)
                        menu-separator
                        (menu-entry 'edit-select-all)
                        (menu-entry 'edit-clear-all))
                  'my-pull-down-menu *screen*
                  :selection-function 'funcall-menu-item))))
    ;; Abstract menu.
    (add-to-menu menu-bar
     (doc-dependent
      (make-menu-item :title "~Abstract"
       :help-string "Abstractions operations"
       :value
       (open-menu (list (menu-entry 'abstract-create)
                        (menu-entry 'abstract-explode)
                        (menu-entry 'abstract-name)
                        (menu-entry 'abstract-unname)
                        menu-separator
                        (menu-entry 'abstract-named-copy)
                        menu-separator
                        (menu-entry 'abstract-remove-current)
                        (menu-entry 'abstract-remove-unused)
                        (menu-entry 'abstract-remove-named-copies))
                  'my-pull-down-menu *screen*
                  :selection-function 'funcall-menu-item))))
    ;; Special Class Menu.
    (setf *booms-icon-class-menu*
          (open-menu nil 'my-pull-down-menu *screen*
                     :selection-function 'funcall-menu-item))
    (add-to-menu menu-bar
     (doc-dependent
      (make-menu-item :title "~Classes"
                      :help-string "Select an icon class"
                      :value *booms-icon-class-menu*)))
    ;; Special Windows Menu.
    (setf *booms-window-menu*
          (open-menu (list (menu-entry 'windows-toggle-palette)
                           (menu-entry 'windows-toggle-status-bar)
                           (menu-entry 'windows-toggle-toolbar)
                           menu-separator)
                     'my-pull-down-menu *screen*
                     :selection-function 'funcall-menu-item))
    (add-to-menu menu-bar
     (make-menu-item :title "~Windows"
                     :help-string "Select a document window / Toggle panes"
                     :value *booms-window-menu*))
    ;; Options menu.
    (add-to-menu menu-bar
     (make-menu-item :title "~Options"
      :help-string "Various BOOMS options"
      :value
      (open-menu (list (setf (first *same-node-same-arg-items*)
                             (menu-entry 'options-same-node-same-arg))
                       (setf (first *dialog-create-items*)
                             (menu-entry 'options-dialog-create))
                       (setf (first *grid-items*)
                             (menu-entry 'options-grid))
                       menu-separator
                       (menu-entry 'options-refresh)
                       (menu-entry 'options-stop-music)
                       (menu-entry 'options-gc))
                 'my-pull-down-menu *screen*
                 :selection-function 'funcall-menu-item)))
    ;; Help menu.
    (add-to-menu menu-bar
     (make-menu-item :title "~Help"
      :help-string "Welcome & About BOOMS"
      :value
      (open-menu (list (menu-entry 'help-welcome)
                       (menu-entry 'help-about))
                 'my-pull-down-menu *screen*
                 :selection-function 'funcall-menu-item)))
    ;; This is an application right-button menu - not in the main menu-bar.
    (setf *app-right-menu*
          (open-menu (list (menu-entry 'file-new)
                           (menu-entry 'file-open)
                           (menu-entry 'file-close-all)
                           (menu-entry 'file-save-all)
                           menu-separator
                           (make-menu-item
                            :title "~Windows"
                            :help-string
                            "Select a document window / Toggle panes"
                            :value *booms-window-menu*)
                           menu-separator
                           (menu-entry 'file-exit))
                     'my-pull-down-menu *screen*
                     :selection-function 'funcall-menu-item))
    ;; This is a document right-button menu - not in the main menu-bar.
    (setf *doc-right-menu*
          (open-menu (list (menu-entry 'file-save)
                           (menu-entry 'file-close)
                           menu-separator
                           (make-menu-item :title "~Classes"
                                           :help-string "Select an icon class"
                                           :value *booms-icon-class-menu*))
                     'my-pull-down-menu *screen*
                     :selection-function 'funcall-menu-item))
    menu-bar))


;;;-------------------------------------------------------------------------;;;
;;;  T O O L B A R  creation                                                ;;;
;;;-------------------------------------------------------------------------;;;

;;; A toolbar class for handling set-focus messages.
(defclass BOOMS-TOOLBAR (toolbar) ())

;;; Refuse focus on toolbar.
(defmethod SET-FOCUS ((w booms-toolbar))
  (set-focus (get-booms-pane)))

;;; Event handling for the toolbar.
(defmethod EVENT ((w booms-toolbar) event buttons data time)
  (select event
    (mouse-in  (say "BOOMS! - The Toolbar    [F12 to toggle]")
               (call-next-method))
    (mouse-out (unsay) (call-next-method))
    (t (call-next-method))))

;;; Show a the help-string when over a button.
(defun SHOW-HELP-TEXT (&rest args)
  (say (help-string (first args))))

;;; Toggle the toolbar.
(defun TOGGLE-TOOLBAR ()
  (if (setf *toolbar-shown* (not *toolbar-shown*))
    (remove-toolbar-from-window *booms-app* *booms-toolbar*)
    (add-toolbar-to-window *booms-app* *booms-toolbar*)))

;;; Toolbar button placement handling - succesively increase offset.
(defun GET-NEXT-TOOLBAR-BUTTON-BOX (&optional seperated)
  (incf *toolbar-button-offset*
        (if seperated
          (+ +toolbar-button-size+ +toolbar-space+)
          +toolbar-button-size+))
  (make-box (- *toolbar-button-offset* +toolbar-button-size+)
            +toolbar-space+
            *toolbar-button-offset*
            (+ +toolbar-button-size+ +toolbar-space+)))

;;; Function to create a toolbar button.
(defun TOOLBAR-BUTTON (command &optional seperatedp)
  (let ((item
         (make-dialog-item
          :widget             'picture-button
          :title              (get command :bmp-name)
          :box                (get-next-toolbar-button-box seperatedp)
          :border             :plain
          :3d-border          t
          :tabstop            nil
          :groupstart         nil
          :background-color   light-gray
          :stretching         nil
          :tooltip            (remove #\~ (get command :title))
          :help-string        (get command :help-string)
          :set-value-fn       #'(lambda (&rest args)
                                  (funcall command)
                                  (get command :toggle-item nil))
          :mouse-in-fn        #'show-help-text
          :mouse-out-fn       #'unsay)))
    (if (get command :doc-dependent nil)
      (doc-dependent item)
      item)))

;;; The toolbar initial width, it will change on startup.
(defconstant +TOOLBAR-INIT-WIDTH+ 500)

;;; Create the toolbar on the given window.
(defun MAKE-BOOMS-TOOLBAR (win)
  (open-dialog
   (list
    (toolbar-button       'file-new              t)
    (toolbar-button       'file-open              )
    (toolbar-button       'file-close             )
    (toolbar-button       'file-save              )
    (toolbar-button       'edit-edit-icon        t)
    (toolbar-button       'edit-cut              t)
    (toolbar-button       'edit-copy              )
    (toolbar-button       'abstract-named-copy    )
    (toolbar-button       'edit-paste             )
    (toolbar-button       'abstract-create       t)
    (toolbar-button       'abstract-name          )
    (toolbar-button       'abstract-explode       )
    (setf (rest *same-node-same-arg-items*)
          (toolbar-button 'options-same-node-same-arg t))
    (setf (rest *dialog-create-items*)
          (toolbar-button 'options-dialog-create  ))
    (setf (rest *grid-items*)
          (toolbar-button 'options-grid           ))
    (toolbar-button       'options-refresh       t)
    (toolbar-button       'options-stop-music     )
    (make-dialog-item
     :widget 'horizontal-scroll-bar ;;'trackbar
     :box (make-box (- +toolbar-init-width+ (+ 160 (* 2 +toolbar-space+)))
                    +toolbar-space+
                    (- +toolbar-init-width+ (+ 10 (* 2 +toolbar-space+)))
                    (+ +toolbar-button-size+ +toolbar-space+))
     :value            (round (* +initial-volume+ 1000))
     :range            '(0 1000)
     :tick-frequency   100
     :value-range-p    t
     :value-range      `(,(round (* +initial-volume+ 1000))
                         ,(round (* +initial-volume+ 1000)))
     :tick-side        :top
     :orientation      :horizontal
     :tabstop          t
     :groupstart       t
     :direction        :down
     :delayed          nil
     :right-attachment :right
     :left-attachment  :right
     :tooltip          "Global Volume"
     :set-value-fn     #'(lambda (dialog-item newvalue oldvalue)
                           (say "MIDI Volume set to ~a%" (round newvalue 10))
                           (setf (trackbar-value-range dialog-item)
                                 (list newvalue newvalue))
                           (set-volume (/ newvalue 1000.0))
                           t)
     :mouse-in-fn      #'(lambda (&rest args)
                           (say "MIDI Volume = ~a%"
                                (round (dialog-item-value (first args)) 10)))
     :mouse-out-fn   #'unsay)
    (make-dialog-item
     :widget 'horizontal-scroll-bar ;;'trackbar
     :box (make-box (- +toolbar-init-width+ (+ 320 (* 2 +toolbar-space+)))
                    +toolbar-space+
                    (- +toolbar-init-width+ (+ 170 (* 2 +toolbar-space+)))
                    (+ +toolbar-button-size+ +toolbar-space+))
     :value            (round (* +initial-tempo+ 1000))
     :range            '(0 1000)
     :tick-frequency   100
     :value-range-p    t
     :value-range      `(,(round (* +initial-tempo+ 1000))
                         ,(round (* +initial-tempo+ 1000)))
     :tick-side        :top
     :orientation      :horizontal
     :tabstop          t
     :groupstart       t
     :direction        :down
     :delayed          nil
     :right-attachment :right
     :left-attachment  :right
     :tooltip          "Global Tempo"
     :set-value-fn     #'(lambda (dialog-item newvalue oldvalue)
                           (say "MIDI Tempo set to ~a%" (round newvalue 10))
                           (setf (trackbar-value-range dialog-item)
                                 (list newvalue newvalue))
                           (set-tempo (/ newvalue 1000.0))
                           t)
     :mouse-in-fn      #'(lambda (&rest args)
                           (say "MIDI Tempo = ~a%"
                                (round (dialog-item-value (first args)) 10)))
     :mouse-out-fn   #'unsay))
   'booms-toolbar     win
   :pop-up-p          nil
   :user-resizable    nil
   :user-closable     nil
   :user-movable      nil
   :user-shrinkable   nil
   :window-state      :shrunk
   :window-border     :plain
   :left-attachment   :left
   :top-attachment    :top
   :right-attachment  :right
   :bottom-attachment :top
   :window-interior   (make-box 0 0 +toolbar-init-width+ +toolbar-height+)))


;;;-------------------------------------------------------------------------;;;
;;;  S T A T U S - B A R  handling                                          ;;;
;;;-------------------------------------------------------------------------;;;

;;; A status-bar class for handling set-focus messages.
(defclass BOOMS-STATUS-BAR (status-bar) ())

;;; Refuse focus on status-bar.
(defmethod SET-FOCUS ((w booms-status-bar))
  (set-focus (get-booms-pane)))

;;; Event handeling for the status-bar.
(defmethod EVENT ((w booms-status-bar) event buttons data time)
  (select event
    (mouse-in  (say "BOOMS! - The Status-Bar    [F11 to toggle]")
               (call-next-method))
    (mouse-out (unsay) (call-next-method))
    (t (call-next-method))))

;;; Functions to display a message on the status-bar (say! beeps).

(defun SAY (format-string &rest format-args)
  (let ((str (if format-args
               (apply #'format nil format-string format-args)
               format-string)))
    (funcall #'window-message *booms-app* str)
    (setf *last-said* str)))

(defun SAY! (format-string &rest format-args)
  (let ((str (if format-args
               (apply #'format nil format-string format-args)
               format-string)))
    (funcall #'window-warning *booms-app* str)
    (setf *last-said* str)))

;;; Clear status-bar message.
(defun UNSAY (&rest garbage)
  (window-message *booms-app* "")
  (setf *last-said* nil))

;;; Same function that will use the second text box.

(defun SAY2 (format-string &rest format-args)
  (let ((str (if format-args
               (apply #'format nil format-string format-args)
               format-string)))
    (set-dialog-item-value *say2-text* str)))

(defun UNSAY2 (&rest garbage)
  (set-dialog-item-value *say2-text* ""))

;;; Toggle the status-bar.
(defun TOGGLE-STATUS-BAR ()
  (if (setf *status-bar-shown* (not *status-bar-shown*))
    (remove-status-bar-from-window *booms-app*)
    (add-status-bar-to-window *booms-app* *booms-status-bar*)))

;;; Create the status-bar on the given window.
(defun MAKE-BOOMS-STATUS-BAR (win)
  (prog1
      (open-dialog
       (list
        (make-dialog-item
          :widget 'static-text
          :name              :default
          :font              +status-bar-font+
          :border            :none
          :left-attachment   :left
          :top-attachment    :top
          :right-attachment  :right
          :bottom-attachment :bottom
          :box               (make-box 15 5 190 23))
        (setf *say2-text*
         (make-dialog-item
          :widget 'static-text
          :value             ""
          :border            :black
          :font              +status-bar-font+
          :left-attachment   :right
          :top-attachment    :top
          :right-attachment  :right
          :bottom-attachment :bottom
          :box               (make-box 200 5 285 23))))
       'booms-status-bar  win
       :pop-up-p          nil
       :user-resizable    nil
       :user-closable     nil
       :user-movable      nil
       :user-shrinkable   nil
       :user-scrollable   nil
       :window-state      :shrunk
       :window-border     :black
       :left-attachment   :left
       :top-attachment    :bottom
       :right-attachment  :right
       :bottom-attachment :bottom
       :window-interior   (make-box 0 0 300 28))
    (remove-status-bar-from-window win)))


;;;-------------------------------------------------------------------------;;;
;;;  K E Y B O A R D  handling                                              ;;;
;;;-------------------------------------------------------------------------;;;

;;; Shortcut for accelerators.
(defun ADD-KEYBOARD-ACCELERATOR (event-synonym func)
  (add-global-keyboard-accelerator event-synonym
                                   #'(lambda (w) (funcall func))))

;;; Add a shortcut for a document function.
(defun ADD-DOCUMENT-ACELERATOR (event-synonym func)
  (add-global-keyboard-accelerator event-synonym
                                   #'(lambda (w)
                                       (funcall func (active-document)))))

;;; The accelerators used in booms, there are more defined by menu entries.
(defun ADD-KEYBOARD-ACCELERATORS ()
  (add-keyboard-accelerator '(shift-key   pc:vk-delete) 'edit-cut)
  (add-keyboard-accelerator '(control-key pc:vk-insert) 'edit-copy)
  (add-keyboard-accelerator '(shift-key   pc:vk-insert) 'edit-paste)
  (add-keyboard-accelerator '(control-key pc:vk-delete) 'edit-delete)
  (add-keyboard-accelerator '(control-key pc:vk-f4)     'file-close)
  (add-keyboard-accelerator '(control-key pc:vk-tab)    'select-next-doc)
  (mapc #'(lambda (x) (apply #'add-document-acelerator x))
        '(((                      pc:vk-up   ) do-move-up          )
          ((                      pc:vk-down ) do-move-down        )
          ((                      pc:vk-left ) do-move-left        )
          ((                      pc:vk-right) do-move-right       )
          ((            shift-key pc:vk-up   ) do-fast-move-up     )
          ((            shift-key pc:vk-down ) do-fast-move-down   )
          ((            shift-key pc:vk-left ) do-fast-move-left   )
          ((            shift-key pc:vk-right) do-fast-move-right  )
          ((control-key           pc:vk-up   ) do-scroll-up        )
          ((control-key           pc:vk-down ) do-scroll-down      )
          ((control-key           pc:vk-left ) do-scroll-left      )
          ((control-key           pc:vk-right) do-scroll-right     )
          ((control-key shift-key pc:vk-up   ) do-fast-scroll-up   )
          ((control-key shift-key pc:vk-down ) do-fast-scroll-down )
          ((control-key shift-key pc:vk-left ) do-fast-scroll-left )
          ((control-key shift-key pc:vk-right) do-fast-scroll-right))))

;;; This is done first to allow these keys to be used in menu-items.
(defun REMOVE-ALLEGRO-GLOBAL-KEYS ()
  (remove-global-keyboard-accelerator '(control-key #\V))
  (remove-global-keyboard-accelerator '(control-key #\C))
  (remove-global-keyboard-accelerator '(control-key #\X)))


;;;-------------------------------------------------------------------------;;;
;;;  C O M M A N D S  handling                                              ;;;
;;;-------------------------------------------------------------------------;;;

;;; This is used to declare a handler and its parameters.
(defmacro DEFCOMMAND (command arguments &body body/parameters)
  (let ((body body/parameters)
        (parameter-setters nil))
    (while (keywordp (first body))
      (push `(get ',command ,(pop body)) parameter-setters)
      (push (pop body) parameter-setters))
    `(progn (defun ,command ,arguments ,@body)
            (setf ,@(nreverse parameter-setters))
            ',command)))

;;; File menu.

(defcommand FILE-NEW ()
  :title       "~New"
  :bmp-name    "new"
  :help-string "Create a new document"
  :accelerator '(control-key #\N)
  (let ((type
         (cond ((null *booms-doc-types*) nil)
               ((null (rest *booms-doc-types*)) (first *booms-doc-types*))
               (t (dialog-get-list-value "New Document" *booms-doc-types*)))))
    (when type
      (make-document type))))

(defcommand FILE-OPEN ()
  :title       "~Open"
  :bmp-name    "open"
  :help-string "Open a saved document"
  :accelerator '(control-key #\O)
  (let ((old-top-doc (active-document))
        (filename
         (my-ask-user-for-existing-pathname
          "Open File"
          :allowed-types '(("Booms files" . "*.bms") ("All files"   . "*.*"))
          :stream        *booms-app*)))
    (let ((existing (find filename (get-all-documents)
                          :key #'doc-pathname :test #'equal)))
      (if existing
        (select-doc existing)
        ;; Open the document, and if success and old window unchanged & clean -
        ;; close it.
        (when (and (not (null filename)) (doc-load filename)
                   old-top-doc
                   (not (doc-pathname old-top-doc))
                   (not (modified-p old-top-doc)))
          (user-close old-top-doc))))))

(defcommand FILE-CLOSE ()
  :title         "~Close"
  :bmp-name      "close"
  :help-string   "Close the active document"
  :accelerator   '(control-key #\W)
  :doc-dependent t
  (user-close (active-document)))

(defcommand FILE-CLOSE-ALL ()
  :title         "Close All"
  :help-string   "Close all documents"
  :doc-dependent t
  ;; First ask all save questions.
  (let ((*dont-close* t))
    (dolist (doc (get-all-documents))
      (unless (user-close doc)
        (return-from file-close-all nil))))
  ;; Now really close.
  (let ((*dont-ask* t))
    (dolist (doc (get-all-documents))
      (unless (user-close doc)
        (return-from file-close-all nil))))
  t)

(defcommand FILE-SAVE (&optional doc)
  :title         "~Save"
  :bmp-name      "save"
  :help-string   "Save the current document"
  :accelerator   '(control-key #\S)
  :doc-dependent t
  (unless doc
    (setf doc (active-document)))
  (if (doc-pathname doc)
    (doc-save doc)
    (file-save-as doc)))

(defcommand FILE-SAVE-AS (&optional doc)
  :title         "Save ~As"
  :help-string   "Save the document under a new name"
  :doc-dependent t
  (unless doc
    (setf doc (active-document)))
  (let ((file-name
         (my-ask-user-for-new-pathname
          "Save File"
          :allowed-types '(("Booms files" . "*.bms") ("All files"   . "*.*"))
          :stream        *booms-app*
          :initial-name  (title doc))))
    (if file-name
      (progn
        (set-doc-pathname doc (merge-pathnames file-name ".bms"))
        (doc-save doc))
      nil)))

(defcommand FILE-SAVE-ALL ()
  :title         "Save All"
  :help-string   "Save all documents"
  :doc-dependent t
  (let ((result t))
    (dolist (doc (get-all-documents))
      (unless (file-save doc)
        (setf result nil)))
    result))

(defcommand FILE-EXIT ()
  :title       "E~xit"
  :help-string "Exit BOOMS"
  :accelerator '(alt-key #\X)
  (user-close *booms-app*))

;;; Edit menu.

(defcommand EDIT-CUT ()
  :title         "Cu~t"
  :bmp-name      "cut"
  :help-string
    "Cut the selected icons to the BOOMS clipboard (for the current type)"
  :accelerator   '(control-key #\X)
  :doc-dependent t
  (do-cut (active-document)))

(defcommand EDIT-COPY ()
  :title         "~Copy"
  :bmp-name      "copy"
  :help-string
    "Copy the selected icons to the BOOMS clipboard (for the current type)"
  :accelerator   '(control-key #\C)
  :doc-dependent t
  (do-copy (active-document)))

(defcommand EDIT-PASTE ()
  :title         "~Paste"
  :bmp-name      "paste"
  :help-string
    "Paste the icons in the BOOMS clipboard  (for the current type)"
  :accelerator   '(control-key #\V)
  :doc-dependent t
  (do-paste (active-document)))

(defcommand EDIT-DELETE ()
  :title         "~Delete"
  :help-string   "Delete the selected icons"
  :accelerator   pc:vk-delete
  :doc-dependent t
  (do-delete (active-document)))

(defcommand EDIT-EDIT-ICON ()
  :title         "~Edit icon"
  :bmp-name      "edit-icon"
  :help-string   "Edit the current icon"
  :accelerator   pc:vk-enter
  :doc-dependent t
  (do-edit-icon (active-document)))

(defcommand EDIT-GRAB-STYLE ()
  :title         "~Grab style"
  :help-string
    "Grab (store) the current icon style (colors and font for messages)"
  :accelerator   '(control-key #\G)
  :doc-dependent t
  (do-grab-style (active-document)))

(defcommand EDIT-APPLY-STYLE ()
  :title         "Apply ~Style"
  :help-string   "Apply the current grabbed style to the selected icons"
  :accelerator   '(control-key #\P)
  :doc-dependent t
  (do-apply-style (active-document)))

(defcommand EDIT-SELECT-ALL ()
  :title         "Select ~All"
  :help-string   "Select all icons in this document"
  :accelerator   '(control-key #\A)
  :doc-dependent t
  (do-select-all (active-document)))

(defcommand EDIT-CLEAR-ALL ()
  :title         "C~lear All"
  :help-string   "Delete all icons in this document"
  :doc-dependent t
  (when (and (icon-list (active-document))
             (my-yes-or-no-p))
    (do-clear-all (active-document))))

;;; Abstract menu.

(defcommand ABSTRACT-CREATE ()
  :title         "Create ~Icon"
  :bmp-name      "abstract"
  :help-string   "Iconify the selected icons to a single abstraction"
  :accelerator   '(control-key #\I)
  :doc-dependent t
  (do-abstract (active-document)))

(defcommand ABSTRACT-EXPLODE ()
  :title         "~Explode"
  :bmp-name      "explode"
  :help-string
    "Explode the selected abstraction[s] to the original components"
  :accelerator   '(control-key #\E)
  :doc-dependent t
  (do-explode (active-document)))

(defcommand ABSTRACT-NAME ()
  :title         "~Name Abstract"
  :bmp-name      "name-abstract"
  :help-string
    (format nil "Name the selected abstraction, if multiple selection - ~
                 create an abstraction and then name it.")
  :accelerator   '(control-key #\M)
  :doc-dependent t
  (do-name-abstraction (active-document)))

(defcommand ABSTRACT-UNNAME ()
  :title         "~Unname"
  :help-string   "Unname the selected abstraction[s]"
  :doc-dependent t
  (do-unname-abstraction (active-document)))

(defcommand ABSTRACT-NAMED-COPY ()
  :title         "Named ~Copy"
  :bmp-name      "named-copy"
  :help-string   "Create a named copy"
  :accelerator   '(control-key shift-key #\C)
  :doc-dependent t
  (do-create-named-copy (active-document)))

(defcommand ABSTRACT-REMOVE-CURRENT ()
  :title         "~Remove Current"
  :help-string   (format nil "Remove current user-defined abstractions or ~
                              user-named-copy from the palette / menu")
  :doc-dependent t
  (do-remove-current-user-node (active-document)))

(defcommand ABSTRACT-REMOVE-UNUSED ()
  :title         "Remove Unu~sed"
  :help-string
    "Remove unused abstractions from the palette / menu for this type"
  :doc-dependent t
  (do-remove-unused-abstractions (active-document)))

(defcommand ABSTRACT-REMOVE-NAMED-COPIES ()
  :title         "Remove Co~pies"
  :help-string
    "Remove all named copies from the palette / menu for this type"
  :doc-dependent t
  (do-remove-named-copies (active-document)))

;;; Windows menu.  (First items, rest is change in runtime).

(defcommand WINDOWS-TOGGLE-PALETTE ()
  :title       "Toggle ~Palette"
  :help-string "Toggle the icon classes palette"
  :accelerator '(pc:vk-f10)
  (toggle-palette))

(defcommand WINDOWS-TOGGLE-STATUS-BAR ()
  :title       "Toggle ~Status-bar"
  :help-string "Toggle the BOOMS Status-bar"
  :accelerator '(pc:vk-f11)
  (toggle-status-bar))

(defcommand WINDOWS-TOGGLE-TOOLBAR ()
  :title       "Toggle ~Toolbar"
  :help-string "Toggle the BOOMS Toolbar"
  :accelerator '(pc:vk-f12)
  (toggle-toolbar))

;;; Options menu.

;;; The next three are a mode-toggle, so it should handle cases where it is
;;; called from a menu-item selection or a toolbar button.

(defcommand OPTIONS-SAME-NODE-SAME-ARG ()
  :title         "~Same-node-same-arg"
  :bmp-name      "node-args"
  :help-string   "Toggle the same-node-same-arg option"
  :accelerator   '(shift-key control-key #\D)
  :doc-dependent t
  :toggle-item   t
  (unless *called-programmatically*
    (let ((*called-programmatically* t)
          (same-node-same-arg-p
           (do-toggle-same-node-same-arg (active-document))))
      (set-mode-items-value *same-node-same-arg-items* same-node-same-arg-p)
      same-node-same-arg-p)))

(defcommand OPTIONS-DIALOG-CREATE ()
  :title         "~Dialog create"
  :bmp-name      "dialog-create"
  :help-string   "Toggle creating an icons with a dialog"
  :accelerator   '(shift-key control-key #\D)
  :doc-dependent t
  :toggle-item   t
  (unless *called-programmatically*
    (let ((*called-programmatically* t)
          (dialog-create-p (do-toggle-dialog-create (active-document))))
      (set-mode-items-value *dialog-create-items* dialog-create-p)
      dialog-create-p)))

(defcommand OPTIONS-GRID ()
  :title         "~Grid"
  :bmp-name      "grid"
  :help-string   "Toggle grid placement for icons"
  :accelerator   '(shift-key control-key #\G)
  :doc-dependent t
  :toggle-item   t
  (unless *called-programmatically*
    (let ((*called-programmatically* t)
          (gridp (do-toggle-grid (active-document))))
      (set-mode-items-value *grid-items* gridp)
      gridp)))

(defcommand OPTIONS-REFRESH ()
  :title         "~Refresh"
  :bmp-name      "refresh"
  :help-string   "Refresh the screen"
  :accelerator   '(control-key #\R)
  :doc-dependent t
  (reset-view (active-document)))

(defcommand OPTIONS-STOP-MUSIC ()
  :title       "Stop ~Music"
  :bmp-name    "stop-music"
  :help-string "Stop playing midi music"
  (midi-reset))

(defcommand OPTIONS-GC ()
  :title       "Garbage ~Collect"
  :help-string "Run a Lisp garbage-collection"
  (gc 2))

;;; Help menu.

(defcommand HELP-WELCOME ()
  :title       "~Welcome"
  :help-string "Hello from BOOMS"
  (popup-welcome-dialog))

(defcommand HELP-ABOUT ()
  :title       "~About BOOMS"
  :help-string "What's BOOMS anyway?"
  (popup-about-dialog))


;;;-------------------------------------------------------------------------;;;
;;;  B O O M S - W I N D O W  class                                         ;;;
;;;-------------------------------------------------------------------------;;;

;;; Defined for handling events.

(defclass BOOMS-WINDOW (non-refreshing-window) ())

(defclass BOOMS-PANE   (non-refreshing-pane)   ())

(defmethod DEFAULT-PANE-CLASS ((stream booms-window)) 'booms-pane)

;;; Message on entering.
(defmethod EVENT ((win booms-window) event buttons data time)
  (select event
    (mouse-in  (say "BOOMS!") (call-next-method))
    (mouse-out (unsay) (call-next-method))
    (t (call-next-method))))

;;; Window menu on a right-click.
(defmethod EVENT ((pane booms-pane) event buttons data time)
  (select event
    (mouse-right-down
     (pop-up-menu *app-right-menu* pane data :center :center :right))
    (t (call-next-method))))

;;; When user closes the window, verify files are saves.
(defmethod USER-CLOSE ((win booms-window))
  (if (file-close-all)
    (progn (call-next-method)
           (terminate-booms)
           (if *dont-throw-done* t (throw :done t)))
    nil))

;;; Make my menus echo their help-string.
(defmethod MENU-ITEM-HIGHLIGHTED ((parent frame-window) (menu my-menu)
                                  (item menu-item))
  (let ((helpstr (menu-item-help-string item)))
    (if helpstr
      (say helpstr)
      (unsay))))


;;;-------------------------------------------------------------------------;;;
;;;  R U N N I N G   B O O M S !                                            ;;;
;;;-------------------------------------------------------------------------;;;

;;; Initialization of global variables.
(defun INITIALIZE-BOOMS-VARIABLES ()
  (setf *booms-app*                   nil
        *booms-menu*                  nil
        *booms-toolbar*               nil
        *booms-status-bar*            nil
        *toolbar-shown*               nil
        *status-bar-shown*            nil
        *misc-menus*                  nil
        *app-right-menu*              nil
        *doc-right-menu*              nil
        *booms-clipboard*             nil
        *booms-doc-types*             nil
        *registering-type*            nil
        *current-type*                nil
        *last-class-for-types*        nil
        *untitled-counter*            0
        *node-classes*                nil
        *current-node-class*          nil
        *booms-icon-class-menu*       nil
        *booms-icon-class-outline*    nil
        *booms-icon-class-menu-items* nil
        *doc-dependent-items*         nil
        *booms-doc-n-list*            nil
        *grabbed-style*               nil
        *last-said*                   nil
        *say2-text*                   nil
        *called-programmatically*     nil
        *dont-close*                  nil
        *dont-ask*                    nil
        *dont-throw-done*             t
        *loading-now*                 nil
        *prevent-next-ready-message*  nil
        *same-node-same-arg-items*    (cons nil nil)
        *dialog-create-items*         (cons nil nil)
        *grid-items*                  (cons nil nil)
        *booms-window-menu*           nil
        *toolbar-button-offset*       3
        *win-top*                     (- (get-caption-height))
        *win-left*                    (- (get-caption-height))
        *booms-icons*                 (make-hash-table)
        *abstraction-exploding*       nil)
  (initialize-doc-vars))

;;; Initialization.
(defun INITIALIZE-BOOMS ()
  (initialize-class-palette)
  (initialize-dialogs)
  (initialize-icon-dialog)
  (initialize-icon-menu)
  (initialize-iconic-leaf)
  (initialize-abstractions)
  (initialize-color-edit)
  (initialize-shared-classes)
  (initialize-music-package)
  (initialize-arithmetic-package)
  (initialize-gcalc-package)
  (midi-reset)
  (set-volume +initial-volume+)
  (set-tempo +initial-tempo+)
  (add-keyboard-accelerators)
  ;; This is needed.
  (setf *registering-type* nil)
  ;; This is for the doc-dependent items.
  (last-doc-removed))

;;; Termination.
(defun TERMINATE-BOOMS ()
  (menu-close (list *booms-menu* *misc-menus*))
  (midi-stop)
  (setf *booms-app* nil))

;;; Create the BOOMS frame window.
(defun MAKE-BOOMS-WINDOW ()
  (open-stream 'booms-window *screen* :io :title "BOOMS!"
               :background-color +booms-back-color+
               :user-scrollable nil
               :window-exterior (make-box 10 10
                                          (- (get-screen-size-x) 10)
                                          (- (get-screen-size-y) 10))))

;;; Start the BOOMS application.
(defun BOOMS (&optional no-logo)
  (let ((*package* (find-package :booms-app))
        (*no-network-errors* t))
    (if *booms-app*
      (select-window *booms-app*)
      (progn
        (initialize-booms-variables)
        (remove-allegro-global-keys)
        (setf *booms-app*        (make-booms-window))
        (setf *booms-menu*       (make-booms-menu *booms-app*))
        (setf *booms-toolbar*    (make-booms-toolbar *booms-app*))
        (setf *booms-status-bar* (make-booms-status-bar *booms-app*))
        (setf (window-icon *booms-app*) (booms-icon 0))
        (set-window-menu          *booms-app* *booms-menu*)
        (add-toolbar-to-window    *booms-app* *booms-toolbar*)
        (add-status-bar-to-window *booms-app* *booms-status-bar*)
        (initialize-booms)
        (unless no-logo (help-welcome))
        (show-palette)
        (file-new)
        ;; This is done for activation with a normal image.
        (let ((*dont-throw-done* nil))
          (catch :done (process-events)))
        (values)))))


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :BOOMSAPP)

