;;;                 -*- Mode: Lisp; Package: BOOMS-APP -*-                  ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         boomsdoc.lsp                                              ;;;
;;; Description:  BOOMS document definition                                 ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Idea by:      Omri Weisman - omri@cs.bgu.ac.il                          ;;;
;;;               Cohavit Taboch  - taboch@cs.bgu.ac.il                     ;;;
;;; Date:         29-12-96                                                  ;;;
;;; Package:      BOOMS-APP                                                 ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  B O O M S D O C  file                                                  ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :booms-app)


;;;-------------------------------------------------------------------------;;;
;;;  D O C - F R A M E  class                                               ;;;
;;;-------------------------------------------------------------------------;;;

(defclass DOC-FRAME (bitmap-window)
  ((pane-class :accessor pane-class :initarg pane-class)))

(defmethod DEFAULT-PANE-CLASS ((stream doc-frame))
  (pane-class stream))

;;; Get the document.
(defmethod DOC ((doc-frame doc-frame))
  (front-window doc-frame))

;;; Inform document when get/lose focus.
(defmethod SET-FOCUS ((doc-frame doc-frame))
  (let ((doc (doc doc-frame)))
    (redisplay-window doc)
    (document-got-focused doc))
  (call-next-method))

;;; Close the document and then the window (unless *dont-close* is true).
(defmethod USER-CLOSE ((doc-frame doc-frame))
  (let ((doc (doc doc-frame)))
    (and (close-doc doc)
         (or *dont-close*
             (prog1 (call-next-method)
               (document-removed doc))))))

;;; Event handling for doc-frame.
(defmethod EVENT ((doc-frame doc-frame) event buttons data time)
  (select event
    (mouse-in  (if *prevent-next-ready-message*
                 (setf *prevent-next-ready-message* nil)
                 (say "Ready")))
    (mouse-out (unsay)))
  (call-next-method))


;;;-------------------------------------------------------------------------;;;
;;;  D O C U M E N T  class                                                 ;;;
;;;-------------------------------------------------------------------------;;;

;;; This class represents a document.
(defclass DOCUMENT (bitmap-pane)
  ((type                  :accessor doc-type                 :initform nil)
   (doc-title             :accessor doc-title                :initform nil)
   (pathname              :accessor doc-pathname             :initform nil)
   (modified-p            :reader   modified-p               :initform nil)
   (icons                 :accessor icon-list                :initform nil)
   (redraws               :accessor redraw-list              :initform nil)
   (same-node-same-arg-on :accessor same-node-same-arg-on    :initform t)
   (dialog-create-on      :accessor dialog-create-on         :initform nil)
   (grid-on               :accessor grid-on                  :initform nil)
   (cur-icon              :accessor current-icon             :initform nil)
   (cur-socket            :accessor current-icon-socket      :initform nil)
   (cur-highlighted       :accessor current-icon-highlighted :initform nil)))

;;; Used to prevent hilighting when needed.
(defvar *PREVENT-HILIGHT* nil)

;;; Used to prevent setting the modified bit (when icon-list is only
;;; reorganized).
(defvar *DONT-MODIFY-DOC* nil)

;;; Get the frame.
(defmethod DOC-FRAME ((doc document))
  (window-parent doc))

;;; Title handling.

(defmethod TITLE ((doc document))
  (doc-title doc))

(defmethod (setf TITLE) (newval (doc document))
  (setf (doc-title doc) newval)
  (setf (stream-title (doc-frame doc)) (stream-title doc))
  (document-title-changed doc))

(defmethod STREAM-TITLE ((doc document))
  (format nil "[ ~:(~a~) ]       ~a  ~:[~;*~]"
          (doc-type doc) (doc-title doc) (modified-p doc)))

;;; Update title when modified.
(defmethod (setf MODIFIED-P) (newval (doc document))
  (unless (eql (slot-value doc 'modified-p) newval)
    (setf (slot-value doc 'modified-p) newval)
    (setf (stream-title (doc-frame doc)) (stream-title doc))))

;;; Pathname handling.
(defmethod SET-DOC-PATHNAME ((doc document) newval)
  (setf (doc-pathname doc) newval)
  (setf (title doc) (pathname-to-string newval)))

;;; Offer to save changes when user closes the document, return true if should
;;; proceed with the closing.
(defmethod CLOSE-DOC ((doc document))
  (when (and (not *dont-ask*) (modified-p doc))
    (case (my-pop-up-message-dialog
           *booms-app* "Save Changes?"
           (format nil "Save changes to ~a ?" (title doc))
           warning-icon
           "~Yes" "~No" "~Cancel")
      (1 (unless (file-save doc) (return-from close-doc nil)))
      (3 (return-from close-doc nil))))
  t)

;;; Some frame operations for convinience.

(defmethod SELECT-DOC ((doc document))
  (select-window (doc-frame doc)))

(defmethod USER-CLOSE ((doc document))
  (user-close (doc-frame doc)))

;;; This method is called whenever the document is modified.  (Do nothing on a
;;; non-document objects).

(defmethod SET-MODIFIED ((obj t))
  t)

(defmethod SET-MODIFIED ((doc document))
  (unless (or *dont-modify-doc* (modified-p doc))
    (setf (modified-p doc) t)))

;;; Changing some of the document's attributes cause it to be considered
;;; 'modified'.  These :before methods will do this job.

(defmethod (setf ICON-LIST)        :before (newval (doc document))
  (set-modified doc))

(defmethod (setf SAME-NODE-SAME-ARG-ON) :before (newval (doc document))
  (set-modified doc))

(defmethod (setf DIALOG-CREATE-ON)      :before (newval (doc document))
  (set-modified doc))

(defmethod (setf GRID-ON)               :before (newval (doc document))
  (set-modified doc))

;;; Finding icon by position.
(defmethod FIND-ICON ((doc document) (pos position))
  (let ((icon (find-if #'(lambda (icon) (icon-intersect-p icon pos))
                       (icon-list doc))))
    (values icon (if icon (icon-find-socket icon pos) nil))))

;;; Select a single icon and bring it to the front.
(defmethod SELECT-ONE-ICON ((doc document)
                            &optional (icon (current-icon doc)))
  (when (or (not (selected icon))
            (not (eq icon (first (icon-list doc))))
            (some #'selected (rest (icon-list doc))))
    (let ((*dont-modify-doc* t))
      (setf (icon-list doc) (delete icon (icon-list doc)))
      (unselect-all doc nil)
      (push icon (icon-list doc))
      (icon-select icon)
      (do-redraws doc))))

;;; Getting the list of selected icons.
(defmethod SELECTED-ICON-LIST ((doc document))
  (remove-if-not #'selected (icon-list doc)))

;;; Fix the the size of the document so that all icons will be accessible.
(defmethod FIX-DOC-SIZE ((doc document))
  (let ((offset-x 0) (offset-y 0) (offset nil) (ofs1 #.(make-position 0 0)))
    (dolist (icon (icon-list doc))
      (setf offset-x (min offset-x (box-left (icon-box icon))))
      (setf offset-y (min offset-y (box-top  (icon-box icon)))))
    (setf offset-x (* (floor offset-x +grid-size-x+) +grid-size-x+))
    (setf offset-y (* (floor offset-y +grid-size-y+) +grid-size-y+))
    (when (or (minusp offset-x) (minusp offset-y))
      (nmake-position ofs1 (- offset-x) (- offset-y))
      (setf offset ofs1)
      (dolist (icon (icon-list doc))
        (icon-move-relative icon offset)))
    (let ((page-width  (page-width  doc t))
          (page-height (page-height doc t)))
      (setf offset-x 0 offset-y 0)
      (dolist (icon (icon-list doc))
        (let ((x (box-left (icon-box icon)))
              (y (box-top (icon-box icon))))
          (setf offset-x (max offset-x (+ x (icon-width icon))))
          (setf offset-y (max offset-y (+ y (icon-height icon))))))
      (setf offset-x (max (- offset-x page-width) 0))
      (setf offset-y (max (- offset-y page-height) 0))
      (when (or (plusp offset-x) (plusp offset-y))
        (set-page-size doc (+ offset-x page-width) (+ offset-y page-height))
        (set-scroll-range doc
         (- (+ offset-x page-width) (box-width  (visible-box doc)))
         (- (+ offset-y page-height) (box-height (visible-box doc))))))
    (when (or offset (plusp offset-x) (plusp offset-y))
      (set-modified doc)
      (with-delayed-redraw ((window-parent doc))
        (when (or (plusp offset-x) (plusp offset-y)) (reset-view doc))
        (do-redraws doc)
        (when offset (scroll doc offset)))
      ;; Update scroll-bars.
      (invalidate-window (window-parent doc))
      (update-window (window-parent doc)))))


;;;-------------------------------------------------------------------------;;;
;;;  H I L I G H T  section                                                 ;;;
;;;-------------------------------------------------------------------------;;;

;;; Highlight the current icon.
(defmethod HILIGHT-CURRENT-ICON ((doc document) &optional dont-say)
  (unless *prevent-hilight*
    (multiple-value-bind (icon socket) (find-icon doc (cursor-position doc))
      (unless (and (current-icon-highlighted doc)
                   (eql icon (current-icon doc))
                   (eql socket (current-icon-socket doc)))
        (with-paint-operation (doc po-invert)
          (when (and (current-icon-highlighted doc)
                     (current-icon doc)
                     (not (eql (current-icon doc) icon)))
            (with-line-width (doc +hilight-line-width+)
              (icon-draw-frame (current-icon doc))))
          (when (and (current-icon-highlighted doc)
                     (current-icon doc)
                     (current-icon-socket doc)
                     (not (and (eql (current-icon doc) icon)
                               (eql (current-icon-socket doc) socket))))
            (draw-circle doc
                         (socket-center (current-icon doc)
                                        (current-icon-socket doc))
                         +socket-size+))
          (when (and icon
                     (not (and (eql (current-icon doc) icon)
                               (current-icon-highlighted doc))))
            (with-line-width (doc +hilight-line-width+)
              (icon-draw-frame icon)))
          (when (and icon socket
                     (not (and (eql (current-icon doc) icon)
                               (eql (current-icon-socket doc) socket)
                               (current-icon-highlighted doc))))
            (draw-circle doc (socket-center icon socket)
                         +socket-size+)))
        (unless dont-say
          (say (if icon (icon-describe icon socket) "Ready")))
        (setf (current-icon doc)             icon)
        (setf (current-icon-socket doc)      socket)
        (setf (current-icon-highlighted doc) t)))))

;;; Unhighlight the current icon.
(defmethod UNHILIGHT-CURRENT-ICON ((doc document))
  (when (current-icon-highlighted doc)
    (when (current-icon doc)
      (with-paint-operation (doc po-invert)
        (with-line-width (doc +hilight-line-width+)
          (icon-draw-frame (current-icon doc)))
        (when (current-icon-socket doc)
          (draw-circle doc
                       (socket-center (current-icon doc)
                                      (current-icon-socket doc))
                       +socket-size+))))
    (setf (current-icon-highlighted doc) nil)))

;;; Select/Unselect all icons (internal).

(defmethod SELECT-ALL ((doc document) &optional (redraw t))
  (dolist (icon (icon-list doc))
    (icon-select icon))
  (when redraw (do-redraws doc)))

(defmethod UNSELECT-ALL ((doc document) &optional (redraw t))
  (dolist (icon (icon-list doc))
    (icon-unselect icon))
  (when redraw (do-redraws doc)))


;;;-------------------------------------------------------------------------;;;
;;;  U S E R - I N T E R F A C E  operations                                ;;;
;;;-------------------------------------------------------------------------;;;

;;; Used as the document user-functions lock.
(defvar *PREVENT-ACTIONS* nil)

;;; Select/Unselect all icons.
(defmethod DO-SELECT-ALL ((doc document) &optional (redraw t))
  (with-lock *prevent-actions*
    (select-all doc)))

;;; Copy the selected icons to the current type entry in *booms-clipboard*,
;;; then normalize (move) them around (0,0).
(defmethod DO-COPY ((doc document))
  (with-lock *prevent-actions*
    (when (some #'selected (icon-list doc))
      (let ((icons (copy-nodes (selected-icon-list doc))))
        (setf (getf *booms-clipboard* *current-type*) icons)
        (setf (icons-center-position icons) #.(make-position 0 0))
        (add-clipboard-named-copy doc icons)
        (say "Copy done")))))

;;; Paste the icons in the current type entry in *booms-clipboard* to the doc
;;; center.
(defmethod DO-PASTE ((doc document))
  (with-lock *prevent-actions*
    (when (getf *booms-clipboard* *current-type*)
      (let ((new-icons (copy-nodes (getf *booms-clipboard* *current-type*))))
        (unhilight-current-icon doc)
        (unselect-all doc nil)
        (let ((pos #.(make-position 0 0)))
          (ncopy-position pos (scroll-position doc))
          (incf (position-x pos) (round (visible-box-width  doc) 2))
          (incf (position-y pos) (round (visible-box-height doc) 2))
          (setf (icons-center-position new-icons) pos)
          (dolist (icon new-icons)
            (setf (icon-doc icon) doc)
            (reset-redraw-info icon)
            (icon-select icon)
            (icon-show icon)
            (when (grid-on doc) (icon-to-grid icon))))
        (setf (icon-list doc) (nconc new-icons (icon-list doc)))
        (fix-doc-size doc)
        (do-redraws doc)
        (hilight-current-icon doc)
        (say "Paste done")))))

;;; Delete the selected icons.
(defmethod DO-DELETE ((doc document))
  (with-lock *prevent-actions*
    (when (some #'selected (icon-list doc))
      (unhilight-current-icon doc)
      (dolist (icon (icon-list doc))
        (when (selected icon)
          (icon-lines-hide icon)
          (delete-links icon)
          (icon-hide icon)))
      (do-redraws doc)
      (setf (icon-list doc) (delete-if #'selected (icon-list doc)))
      (hilight-current-icon doc))))

;;; Cut simply copies the icons and then deletes them.
(defmethod DO-CUT ((doc document))
  (with-lock *prevent-actions*
    (let ((*prevent-actions* nil))
      (do-copy doc)
      (do-delete doc))))

;;; Edit the current icon.
(defmethod DO-EDIT-ICON ((doc document))
  (with-lock *prevent-actions*
    (when (= (count-if #'selected (icon-list doc)) 1)
      (icon-edit (find-if #'selected (icon-list doc))))))

;;; Grab the current icon's style.
(defmethod DO-GRAB-STYLE ((doc document))
  (with-lock *prevent-actions*
    (when (= (count-if #'selected (icon-list doc)) 1)
      (setf *grabbed-style* (icon-style (find-if #'selected (icon-list doc))))
      (say "Grabbed."))))

;;; Paint with the previously grabbed style.
(defmethod DO-APPLY-STYLE ((doc document))
  (with-lock *prevent-actions*
    (when (and (some #'selected (icon-list doc)) *grabbed-style*)
      (dolist (icon (icon-list doc))
        (when (selected icon)
          (setf (icon-style icon) *grabbed-style*)))
      (do-redraws doc)
      (say "Done."))))

;;; Delete all icons.
(defmethod DO-CLEAR-ALL ((doc document))
  (with-lock *prevent-actions*
    (when (icon-list doc)
      (unhilight-current-icon doc)
      (setf (icon-list doc) nil)
      (clear-page doc))))

;;; Toggle the same-node-same-arg option.
(defmethod DO-TOGGLE-SAME-NODE-SAME-ARG ((doc document))
  (with-lock *prevent-actions*
    (setf (same-node-same-arg-on doc) (not (same-node-same-arg-on doc)))))

;;; Toggle the dialog-create option.
(defmethod DO-TOGGLE-DIALOG-CREATE ((doc document))
  (with-lock *prevent-actions*
    (setf (dialog-create-on doc) (not (dialog-create-on doc)))))

;;; Toggle the grid option.
(defmethod DO-TOGGLE-GRID ((doc document))
  (with-lock *prevent-actions*
    (when (setf (grid-on doc) (not (grid-on doc)))
      (dolist (icon (icon-list doc))
        (icon-to-grid icon))
      (do-redraws doc))
    (grid-on doc)))

;;; Move the selected icons in a direction.
(defmethod DO-MOVE ((doc document) (pos position))
  (with-lock *prevent-actions*
    (when (and (eql doc (get-focus *booms-app*))
               (some #'selected (icon-list doc)))
      (unhilight-current-icon doc)
      (dolist (icon (icon-list doc))
        (when (selected icon)
          (icon-move-relative icon pos)))
      (fix-doc-size doc)
      (do-redraws doc)
      (hilight-current-icon doc))))

;;; The four direction functions for above + four fast versions.

(defmethod DO-MOVE-UP ((doc null)) nil)
(defmethod DO-MOVE-UP ((doc document))
  (do-move doc #.(make-position 0 (- +grid-size-y+))))

(defmethod DO-MOVE-DOWN ((doc null)) nil)
(defmethod DO-MOVE-DOWN ((doc document))
  (do-move doc #.(make-position 0 +grid-size-y+)))

(defmethod DO-MOVE-LEFT ((doc null)) nil)
(defmethod DO-MOVE-LEFT ((doc document))
  (do-move doc #.(make-position (- +grid-size-x+) 0)))

(defmethod DO-MOVE-RIGHT ((doc null)) nil)
(defmethod DO-MOVE-RIGHT ((doc document))
  (do-move doc #.(make-position +grid-size-x+ 0)))

(defmethod DO-FAST-MOVE-UP ((doc null)) nil)
(defmethod DO-FAST-MOVE-UP ((doc document))
  (do-move doc #.(make-position 0 (- (* +grid-size-y+ 6)))))

(defmethod DO-FAST-MOVE-DOWN ((doc null)) nil)
(defmethod DO-FAST-MOVE-DOWN ((doc document))
  (do-move doc #.(make-position 0 (* +grid-size-y+ 6))))

(defmethod DO-FAST-MOVE-LEFT ((doc null)) nil)
(defmethod DO-FAST-MOVE-LEFT ((doc document))
  (do-move doc #.(make-position (- (* +grid-size-x+ 4)) 0)))

(defmethod DO-FAST-MOVE-RIGHT ((doc null)) nil)
(defmethod DO-FAST-MOVE-RIGHT ((doc document))
  (do-move doc #.(make-position (* +grid-size-x+ 4) 0)))

;;; Move the selected icons in a direction.
(defmethod DO-SCROLL ((doc document) (pos position))
  (with-lock *prevent-actions*
    (when (eql doc (get-focus *booms-app*))
      (let ((scr-pos #.(make-position 0 0)))
        (ncopy-position scr-pos (scroll-position doc))
        (nposition+ scr-pos pos)
        (scroll-to doc scr-pos)))))

;;; The four direction functions for above + four fast versions.

(defmethod DO-SCROLL-UP ((doc null)) nil)
(defmethod DO-SCROLL-UP ((doc document))
  (do-scroll doc #.(make-position 0 (- +grid-size-y+))))

(defmethod DO-SCROLL-DOWN ((doc null)) nil)
(defmethod DO-SCROLL-DOWN ((doc document))
  (do-scroll doc #.(make-position 0 +grid-size-y+)))

(defmethod DO-SCROLL-LEFT ((doc null)) nil)
(defmethod DO-SCROLL-LEFT ((doc document))
  (do-scroll doc #.(make-position (- +grid-size-x+) 0)))

(defmethod DO-SCROLL-RIGHT ((doc null)) nil)
(defmethod DO-SCROLL-RIGHT ((doc document))
  (do-scroll doc #.(make-position +grid-size-x+ 0)))

(defmethod DO-FAST-SCROLL-UP ((doc null)) nil)
(defmethod DO-FAST-SCROLL-UP ((doc document))
  (do-scroll doc #.(make-position 0 (- (* +grid-size-y+ 4)))))

(defmethod DO-FAST-SCROLL-DOWN ((doc null)) nil)
(defmethod DO-FAST-SCROLL-DOWN ((doc document))
  (do-scroll doc #.(make-position 0 (* +grid-size-y+ 4))))

(defmethod DO-FAST-SCROLL-LEFT ((doc null)) nil)
(defmethod DO-FAST-SCROLL-LEFT ((doc document))
  (do-scroll doc #.(make-position (- (* +grid-size-x+ 4)) 0)))

(defmethod DO-FAST-SCROLL-RIGHT ((doc null)) nil)
(defmethod DO-FAST-SCROLL-RIGHT ((doc document))
  (do-scroll doc #.(make-position (* +grid-size-x+ 4) 0)))


;;;-------------------------------------------------------------------------;;;
;;;  E V E N T S  handling                                                  ;;;
;;;-------------------------------------------------------------------------;;;

;;; Event handling for document.
(defmethod EVENT ((doc document) event buttons data time)
  (when (eql event mouse-middle-down) ; Make middle shortcut for double-click.
    (setf event mouse-left-double-click))
  (select event
    (timer-event (return-from event nil))
    (mouse-moved (hilight-current-icon doc)
                 (call-next-method)) ; Needed for setting the cursor.
    (mouse-left-down         (left-click   doc data buttons))
    (mouse-right-down        (right-click  doc data))
    (mouse-left-double-click (double-click doc data))
    (t (call-next-method))))

;;; Many things can be done on a left-click...
(defmethod LEFT-CLICK ((doc document) (pos0 position) buttons)
  (with-lock *prevent-actions*
    (let ((current (current-icon doc))
          (socket  (current-icon-socket doc))
          (pos #.(make-position 0 0))
          (*prevent-hilight* t))
      (ncopy-position pos pos0)
      (unhilight-current-icon doc)
      (cond
        ;; Control+Click - edit the icon.
        ((and current (button-match buttons control-key))
         (icon-edit current socket))
        ;; Shift+Click icon - toggle its selection, if turend on - start drag.
        ((and current (button-match buttons shift-key))
         (when (icon-select-toggle current)
           (do-redraws doc)
           (do-dragging doc pos)))
        ;; Click a socket - start dragging a connection.
        ((and current socket)
         (select-one-icon doc current)
         (do-socket-dragging doc))
        ;; Click icon - drag, if unselected - bring to front + select only it.
        (current
         (unless (selected current)
           (select-one-icon doc current))
         (do-dragging doc pos))
        ;; Clicking on no icon + Shift - box selection toggle.
        ((button-match buttons shift-key)
         (let ((box #.(make-box 0 0 0 0)))
           (nget-box doc box t t pos)
           (normalize-box box)
           (dolist (icon (icon-list doc))
             (when (icon-intersect-p icon box)
               (icon-select-toggle icon)))))
        ;; Clicking on no icon - box selection + bring to front.
        (t
         (let ((box #.(make-box 0 0 0 0))
               (unselected-icons nil)
               (selected-icons nil))
           (nget-box doc box t t pos)
           (normalize-box box)
           (unselect-all doc nil)
           (dolist (icon (reverse (icon-list doc)))
             (if (icon-intersect-p icon box)
               (progn
                 (icon-select icon)
                 (push icon selected-icons))
               (push icon unselected-icons)))
           (let ((*dont-modify-doc* t))
             (setf (icon-list doc)
                   (append selected-icons unselected-icons))))))
      (do-redraws doc))
    (setf (current-icon doc) nil)
    (hilight-current-icon doc)))

;;; Pop up the specialized menu for the current node or the class menu if no
;;; current node.
(defmethod RIGHT-CLICK ((doc document) (pos position))
  (with-lock *prevent-actions*
    (unhilight-current-icon doc)
    (unsay)
    (let ((menu-pos (make-position 0 0))
          (current (current-icon doc))
          (*prevent-hilight* t))
      (ncopy-position menu-pos pos)
      (nposition- menu-pos (scroll-position doc))
      (if current
        (progn
          (select-one-icon doc current)
          (pop-up-icon-menu current doc menu-pos))
        (pop-up-menu *doc-right-menu* doc menu-pos :center :center :right)))
    (setf (current-icon doc) nil)
    ;; If something was said - keep it.
    (hilight-current-icon doc (if *last-said* t nil))
    (redisplay-window doc)
    (setf *prevent-next-ready-message* t)))

;;; Create a new icon or evaluate current one.
(defmethod DOUBLE-CLICK ((doc document) (pos position))
  (with-lock *prevent-actions*
    (if (current-icon doc)
      (progn
        (select-one-icon doc (current-icon doc))
        (icon-eval (current-icon doc)))
      (create-booms-icon doc pos))
    (redisplay-window doc)))

;;; Create a new icon, no-position will be true when an icon is dragged from
;;; the classes palette.  This should handle cases where the icon created is
;;; actually an icon list - created by an instanciation a named copy.
(defmethod CREATE-BOOMS-ICON ((doc document) (pos position)
                              &optional no-position)
  (let* ((dummy    #.(make-box 0 0 0 0))
         (grid-on  (grid-on doc))
         (icon     (make-icon *current-node-class*
                              :centered t :document doc :pos pos))
         (icons    (if (listp icon) icon (list icon)))
         (draw-p nil)
         (draw-fn  (if (listp icon)
                     #'(lambda (win box)
                         (when (setf draw-p (not draw-p))
                           (setf (icons-center-position icons)
                                 (nbox-top-left #.(make-position 0 0) box)))
                         (dolist (i icon)
                           (when grid-on (icon-to-grid i))
                           (icon-draw-frame i)))
                     #'(lambda (win box)
                         (icon-set-pos icon (nnbox-top-left box))
                         (when grid-on (icon-to-grid icon))
                         (icon-draw-frame icon)))))
    (when grid-on
      (dolist (icon icons)
        (icon-to-grid icon)))
    (unless no-position
      (with-events-disabled (doc (window-parent doc) *booms-app*)
        (nmake-box-from-corners dummy pos pos)
        (set-cursor-position doc pos)
        (nget-shape-fixed-box doc dummy draw-fn draw-fn t t)))
    (setf *icon-edit-called* nil)
    (when (icon-initialize icon)
      (unselect-all doc nil)
      (dolist (icon icons)
        (icon-select icon)
        (icon-show icon))
      (setf (icon-list doc) (nconc icons (icon-list doc)))
      (fix-doc-size doc)
      (do-redraws doc)
      (when (and (not (listp icon))
                 (dialog-create-on doc)
                 (not *icon-edit-called*))
        (let ((*prevent-hilight* t))
          (unless (icon-edit icon)
            (icon-hide icon)
            (do-redraws doc)
            (pop (icon-list doc)))))
      (hilight-current-icon doc))))


;;;-------------------------------------------------------------------------;;;
;;;  D R A G G I N G  handling                                              ;;;
;;;-------------------------------------------------------------------------;;;

;;; The icon list we're dragging.
(defvar *DRAGGING-ICONS* nil)
;;; Have we broken the threshold and started dragging?
(defvar *DRAGGING-STARTED* nil)
;;; The dragging position.
(defvar *DRAGGING-POS* nil)

;;; Do the dragging.
(defmethod DO-DRAGGING ((doc document) (pos position))
  (initialize-dragging doc pos)
  (when (button-match (mouse-button-state) left-mouse-button)
    (with-paint-operation (doc po-invert)
      (dolist (icon *dragging-icons*)
        (icon-draw-frame icon)))
    (nget-shape-box doc #.(make-box 0 0 0 0) #'drag-draw #'true t t pos)
    (with-paint-operation (doc po-invert)
      (dolist (icon *dragging-icons*)
        (icon-draw-frame icon))))
  (fix-doc-size doc))

;;; Called before a dragging.
(defmethod INITIALIZE-DRAGGING ((doc document) (pos position))
  (setf *dragging-icons*   (selected-icon-list doc))
  (setf *dragging-started* nil)
  (setf *dragging-pos*     (ncopy-position #.(make-position 0 0) pos)))

;;; Called to draw the dragged icons.
(defun DRAG-DRAW (doc box)
  (when (grid-on doc)
    (set-box-right  box
                    (+ (* (round (box-width  box) +grid-size-x+) +grid-size-x+)
                       (box-left box)))
    (set-box-bottom box
                    (+ (* (round (box-height box) +grid-size-y+) +grid-size-y+)
                       (box-top  box))))
  (when (button-match (mouse-button-state) shift-key)
    (if (< (abs (box-width box)) (abs (box-height box)))
      (set-box-right box (box-left box))
      (set-box-bottom box (box-top box))))
  (let ((offset #.(make-position 0 0)))
    (nbox-bottom-right offset box)
    (nposition- offset *dragging-pos*)
    (when (and (not
                (and (zerop (position-x offset)) (zerop (position-y offset))))
               (or *dragging-started*
                   (> (position-length offset) +drag-threshold+)))
      (setf *dragging-started* t)
      (with-delayed-redraw (doc)
        (dolist (icon *dragging-icons*)
          (icon-draw-frame icon)
          (icon-move-relative icon offset)
          (icon-draw-frame icon)))
      (redisplay-window doc)
      (nbox-bottom-right *dragging-pos* box))))

;;; Drag a socket connection.
(defmethod DO-SOCKET-DRAGGING ((doc document))
  (let ((first-icon        (current-icon doc))
        (first-socket      (current-icon-socket doc))
        second-icon        second-socket
        (disconnected      nil)
        (*prevent-hilight* nil)
        (pos               #.(make-position 0 0)))
    (when (and (integerp first-socket)
               (not (and (list-args-p first-icon)
                         (= first-socket (length (arguments first-icon)))))
               (iconp (getarg0 first-icon first-socket)))
      (setf disconnected (list first-icon first-socket))
      (setf first-icon (getarg0 first-icon first-socket))
      (setf first-socket :top)
      (delarg0 (first disconnected) (second disconnected))
      (setarg0 (first disconnected) (second disconnected) +no-value+)
      (do-redraws doc))
    (when (button-match (mouse-button-state) left-mouse-button)
      (nget-shape-box doc #.(make-box 0 0 0 0)
                      #'socket-line-draw #'socket-line-draw t t
                      (ncopy-position
                       pos
                       (socket-center first-icon first-socket)))
      (setf second-icon   (current-icon doc))
      (setf second-socket (current-icon-socket doc))
      (unhilight-current-icon doc)
      (when (and second-icon (not second-socket))
        (setf second-socket
              (cond
                ((integerp first-socket) :top)
                ((and disconnected (eql second-icon (first disconnected)))
                 (second disconnected))
                (t (default-socket-number second-icon first-icon)))))
      (cond
        ((and (integerp first-socket) (eql second-socket :top))
         (setarg0 first-icon first-socket second-icon))
        ((and (integerp second-socket) (eql first-socket :top))
         (setarg0 second-icon second-socket first-icon)))
      (when (and disconnected (list-args-p (first disconnected)))
        (delarg0 (first disconnected)
                 (if (and (eql (first disconnected) second-icon)
                          (integerp second-socket)
                          (>= (second disconnected) second-socket))
                   (1+ (second disconnected))
                   (second disconnected)))))))

;;; Draw the socket line - do hilighting meanwhile.
(defmethod SOCKET-LINE-DRAW ((doc document) (box box))
  (hilight-current-icon doc)
  (draw-line doc (nnbox-top-left box) (nnbox-bottom-right box)))


;;;-------------------------------------------------------------------------;;;
;;;  R E D R A W  handling                                                  ;;;
;;;-------------------------------------------------------------------------;;;
;;; Redrawing is done using request-icon-redraw which is used to schedule an
;;; icon redraw, then, do-redraws will perform the actual redrawing.  This way,
;;; the doc can accumulate redraw events and handle them alltogather more
;;; effectively.  The redraw-info is kept in the icons, this way several
;;; redraws for a single icon can be reduced into one.

;;; This is called maximum once per redraw for a given icon and will add this
;;; icon to the redraw-list of the doc.
(defmethod REQUEST-ICON-REDRAW ((doc document) icon)
  (push icon (redraw-list doc)))

;;; This will do all necessary redraws.
(defmethod DO-REDRAWS ((doc document))
  (when (and (icon-list doc) (redraw-list doc))
    (with-delayed-redraw (doc)
      (dolist (icon (icon-list doc))
        (icon-lines-hide icon))
      (if (< (/ (length (redraw-list doc)) (length (icon-list doc)))
             +redraw-all-ratio-threshold+)
        (let ((r-icon-list (reverse (icon-list doc))))
          (dolist (icon (redraw-list doc))
            (when (first (redraw-info icon))
              (do-icon-erase-old icon)
              (dolist (box (icon-old-box-list icon))
                (with-clipping (doc box)
                  (dolist (i r-icon-list)
                    (when (and (shown i) (icon-intersect-p i box)
                               (not (eql i icon)))
                      (icon-draw i))))))
            (when (rest (redraw-info icon))
              (do-icon-draw icon)
              (dolist (box (icon-box-list icon))
                (with-clipping (doc box)
                  (dolist (i (rest (member icon r-icon-list)))
                    (when (icon-intersect-p i box)
                      (icon-draw i))))))))
        (progn
          (dolist (icon (redraw-list doc))
            (do-icon-erase-old icon))
          (dolist (icon (reverse (icon-list doc)))
            (icon-reshow icon)
            (do-icon-draw icon))))
      (dolist (icon (icon-list doc)) (icon-lines-show icon))
      (setf (redraw-list doc) nil))
    (redisplay-window doc)))

;;; Erase all view, then reshow everything (do-redraw will do that).
(defmethod RESET-VIEW ((doc document))
  (with-delayed-redraw (doc)
    (unhilight-current-icon doc)
    (dolist (icon (icon-list doc))
      (icon-lines-hide icon)
      (icon-reshow icon))
    (clear-page doc))
  (do-redraws doc))


;;;-------------------------------------------------------------------------;;;
;;;  D O C U M E N T  constructor                                           ;;;
;;;-------------------------------------------------------------------------;;;

;;; These two methods will produce the class types for the various document
;;; types.  The default give the general class types.

(defmethod FRAME-TYPE    (type) 'doc-frame)

(defmethod PANE-TYPE     (type) 'document)

(defmethod DOCUMENT-ICON (type) (booms-icon 1))

;;; Create a new document - create the frame window and connect objects, the
;;; document will be a pane subclass.
(defun MAKE-DOCUMENT (type
                      &rest initargs
                      &key (parent   (get-booms-pane))
                           (title    (get-next-untitled-title type))
                           (pathname nil)
                           (exterior (get-new-exterior parent))
                           &allow-other-keys)
  (remf initargs :parent)
  (remf initargs :title)
  (remf initargs :pathname)
  (remf initargs :exterior)
  (let* ((doc-frame (apply #'open-stream
                           (frame-type type) parent :io
                           :title            title
                           :background-color +doc-back-color+
                           :window-exterior  exterior
                           :pane-class       (pane-type type)
                           initargs))
         (doc (and doc-frame (doc doc-frame))))
    (if doc-frame
      (progn (setf (doc-type doc) type)
             (setf (title doc) title)
             (when pathname (set-doc-pathname doc pathname))
             (set-window-icon doc-frame (document-icon type))
             (set-cursor doc-frame arrow-cursor)
             (document-added doc))
      (my-pop-up-message-dialog
       parent "Booms Error" "Cannot create more windows" warning-icon "~OK"))
    doc))


;;;-------------------------------------------------------------------------;;;
;;;  D O C U M E N T  variables initialization                              ;;;
;;;-------------------------------------------------------------------------;;;

(defun INITIALIZE-DOC-VARS ()
  (setf *prevent-hilight*  nil
        *dont-modify-doc*  nil
        *prevent-actions*  nil
        *dragging-icons*   nil
        *dragging-started* nil
        *dragging-pos*     nil))


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :BOOMSDOC)

