;;;                  -*- Mode: Lisp; Package: MIDI-LIB -*-                  ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         midi-lib.lsp                                              ;;;
;;; Description:  BOOMS C function FFI interface to MIDI                    ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Date:         24-12-96                                                  ;;;
;;; Package:      MIDI-LIB                                                  ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  M I D I - L I B  file                                                  ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :midi-lib)


;;;-------------------------------------------------------------------------;;;
;;;  C  declerations and constants                                          ;;;
;;;-------------------------------------------------------------------------;;;

(defvar *CTYPES-DONE* nil)

(when (not *ctypes-done*)
  (defctype -void   :void)
  (defctype -bool   :short-bool)
  (defctype -uint   :unsigned-short)
  (defctype -word   :unsigned-short)
  (defctype -dword  :unsigned-long)
  (defctype -char   :char)
  (defctype -lpstr  (-char **))
  (defctype -lpcstr -lpstr)
  (defctype -lpcstr -lpstr)
  (setf *ctypes-done* t))


;;;-------------------------------------------------------------------------;;;
;;;  M I D I - L I B  C-functions interface                                 ;;;
;;;-------------------------------------------------------------------------;;;

(defconstant MIDI-LIB-DLL "midi-lib.dll")

;;; SoundName should be the name of an available wave file (or system sound
;;; event name).
(defun-dll PLAY-SOUND ((SoundName -lpcstr))
  :return-type  -bool
  :library-name midi-lib-dll
  :entry-name   "play_sound"
  :call-mode    :c)

;;; Is the MIDI channel open?
(defun-dll MIDI-OPEN-P ()
  :return-type  -bool
  :library-name midi-lib-dll
  :entry-name   "midi_open_p"
  :call-mode    :c)

;;; Open the MIDI channel.
(defun-dll MIDI-OPEN ()
  :return-type  -bool
  :library-name midi-lib-dll
  :entry-name   "midi_open"
  :call-mode    :c)

;;; Close the MIDI channel.
(defun-dll MIDI-CLOSE ()
  :return-type  -bool
  :library-name midi-lib-dll
  :entry-name   "midi_close"
  :call-mode    :c)

;;; Reset (close/open) the MIDI channel.
(defun-dll MIDI-RESET ()
  :return-type  -bool
  :library-name midi-lib-dll
  :entry-name   "midi_reset"
  :call-mode    :c)

;;; Set the MIDI channel volume (0-65535).  (Set two sides to the same volume).
(defun-dll SET-VOLUME-C ((volume -word))
  :return-type  -bool
  :library-name midi-lib-dll
  :entry-name   "set_volume"
  :call-mode    :c)

;;; Get the MIDI volume.
(defun-dll GET-VOLUME-C ()
  :return-type  -word
  :library-name midi-lib-dll
  :entry-name   "get_volume"
  :call-mode    :c)

;;; Play midi node 'pitch' (0-127) on a velocity of 'velocity' (0-127),
;;; starting from time 'start' for a duration of 'duration' (both msec).
(defun-dll PLAY-NOTE-C ((start -dword) (duration -uint)
                        (pitch -char)  (velocity -char))
  :return-type  -void
  :library-name midi-lib-dll
  :entry-name   "play_note"
  :call-mode    :c)

;;; Get current Windows time for the play-note function.
(defun-dll GET-TIME ()
  :return-type  -dword
  :library-name midi-lib-dll
  :entry-name   "get_time"
  :call-mode    :c)

;;; Low-level function - put a MIDI 'message' event to occur at time 'time' in
;;; the internal MIDI-LIB buffer.
(defun-dll PUT-EVENT ((message -dword) (time -dword))
  :return-type  -void
  :library-name midi-lib-dll
  :entry-name   "put_event"
  :call-mode    :c)

;;; Send a MIDI message now.
(defun-dll MIDI-SEND ((message -dword))
  :return-type  -bool
  :library-name midi-lib-dll
  :entry-name   "midi_send"
  :call-mode    :c)


;;;-------------------------------------------------------------------------;;;
;;;  M I D I - L I B  Lisp functions interface                              ;;;
;;;-------------------------------------------------------------------------;;;

;;; Open the MIDI channel.  Midi-start shouldn't be used since it is done
;;; automatically on the DLL startup.
(defun MIDI-START ()
  (midi-open))

;;; Stop the MIDI channel and unlink the DLL.
(defun MIDI-STOP ()
  (midi-close)
  (unlink-dll midi-lib-dll))

;;; The global tick frequency - the length in miliseconds of a whole note.
(defvar *GLOBAL-TICK-FREQUENCY* 1000.0)

;;; The current play session start time.  This should be set before using
;;; play-note for default start-time..
(defvar *PLAY-START-TIME* nil)

;;; Some time to delay to let the calculations finish.
(defconstant +DELAY-BEFORE-PLAY+ 20)

;;; Call this before starting to play - to set *play-start-time*.
(defun START-PLAY ()
  (setf *play-start-time* (+ (get-time) +delay-before-play+)))

;;; Call this after finishing playing - to unset *play-start-time*.
(defun END-PLAY ()
  (setf *play-start-time* nil))

;;; A Lisp interface to the play-note C function.
(defun PLAY-NOTE (pitch &key (start (or *play-start-time* (get-time)))
                             (wait 0) (duration 1.0) (velocity 1.0))
  (incf start (round (* wait *global-tick-frequency*)))
  (play-note-c start (round (* duration *global-tick-frequency*))
               pitch (round (* velocity 127))))

;;; Set the volume, use a value between 0 and 1.
(defun SET-VOLUME (volume)
  (set-volume-c (round (* volume 65535))))

;;; Return the volume as a number from 0 to 1.
(defun GET-VOLUME ()
  (/ (get-volume-c) 65535))

;;; Set the tempo, use a value between 0 and 1.
(defun SET-TEMPO (tempo)
  (setf *global-tick-frequency* (+ 50 (* tempo 2000.0))))

;;; Return the tempo as a number from 0 to 1.
(defun GET-TEMPO ()
  (/ (- *global-tick-frequency* 50) 2000.0))


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :MIDI-LIB)

