;;;                 -*- Mode: Lisp; Package: BOOMS-APP -*-                  ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         shared-i.lsp                                              ;;;
;;; Description:  BOOMS shared icon types                                   ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Date:         28-12-96                                                  ;;;
;;; Package:      BOOMS-APP                                                 ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  S H A R E D - I  file                                                  ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :booms-app)


;;;-------------------------------------------------------------------------;;;
;;;  I C O N I C - I D E N T I T Y  class                                   ;;;
;;;-------------------------------------------------------------------------;;;

;;; The iconic-identity class, named identity, nothing to do here.

(defclass ICONIC-IDENTITY (iconic-node identity) ())

(set-booms-name iconic-identity 'identity)

(defmethod ICON-ICON ((icon iconic-identity)) (booms-icon 23))

;;; Use this method for implementing document-specific identity icons.  The
;;; default returns the usual icon.
(defmethod ICONIC-IDENTITY-CLASS-NAME ((doctype symbol))
  'iconic-identity)


;;;-------------------------------------------------------------------------;;;
;;;  I C O N I C - L I S P - V A L U E  class                               ;;;
;;;-------------------------------------------------------------------------;;;

;;; This is a general Leaf-value leaf node.

(defleaf ICONIC-LEAF-VALUE (leaf-value) ())

(set-booms-name iconic-leaf-value 'leaf-value)

(defmethod ICON-ICON ((icon leaf-value)) (booms-icon 24))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N I C - T E X T  class                                           ;;;
;;;-------------------------------------------------------------------------;;;

;;; This class is used to put short text messages on the document.

(defclass ICONIC-TEXT (icon)
  ((text       :initform ""
               :initarg :icon-text       :reader   icon-text)
   (text-font  :initform +icon-text-font+
               :initarg :text-font       :reader   text-font)
   (text-color :initform +icon-text-color+
               :initarg :text-color      :reader   text-color)
   (adjust-p   :initform t
               :initarg :should-adjust-p :accessor should-adjust-p)))

(set-booms-name iconic-text 'text)

;;; Call this for readjusting the icon-box on the next draw.
(defmethod ICON-SHOULD-ADJUST ((icon iconic-text))
  (setf (should-adjust-p icon) t)
  (copy-boxes-to-old-boxes icon)
  (when (shown icon)
    (icon-hide icon)
    (icon-show icon)))

;;; When setting the text string - should adjust.
(defmethod (setf ICON-TEXT) (newval (icon iconic-text))
  (unless (equal newval (slot-value icon 'text))
    (setf (slot-value icon 'text) newval)
    (icon-should-adjust icon)
    (set-modified (icon-doc icon))))

;;; When setting the text font - should adjust.
(defmethod (setf TEXT-FONT) (newval (icon iconic-text))
  (unless (equal newval (slot-value icon 'text-font))
    (setf (slot-value icon 'text-font) newval)
    (icon-should-adjust icon)
    (set-modified (icon-doc icon))))

;;; When setting the text color - should adjust.
(defmethod (setf TEXT-COLOR) (newval (icon iconic-text))
  (unless (equal newval (slot-value icon 'text-color))
    (setf (slot-value icon 'text-color) newval)
    (icon-should-adjust icon)
    (set-modified (icon-doc icon))))

;;; Set the extra field on a copy.
(defmethod MAKE-COPY ((icon iconic-text))
  (let ((result (call-next-method)))
    (setf (icon-text  result) (icon-text  icon)
          (text-font  result) (text-font  icon)
          (text-color result) (text-color icon))
    result))

;;; Save these slots.
(defmethod LOAD-SLOTS append ((x iconic-text))
  '(text text-font text-color))

;;; Adjust the box using the string size.
(defmethod ADJUST-ICON-BOX ((icon iconic-text))
  (setf (should-adjust-p icon) nil)
  (let* ((doc    (icon-doc icon))
         (box    (icon-box icon))
         (pos    (icon-center-pos icon))
         (str    (icon-text icon)))
    (with-font (doc (text-font icon))
      (let ((width  (+ (stream-string-width doc str) #.(* +icon-margin+ 2)))
            (height (+ (line-height doc) #.(* +icon-margin+ 2))))
        (nmake-box-relative box
                            (- (position-x pos) (round width 2))
                            (- (position-y pos) (round height 2))
                            width height)))))

;;; Ask user for initial value using the text pane.
(defmethod ICON-INITIALIZE ((icon iconic-text))
  (setf (icon-text icon) "")
  (icon-edit icon nil :text t))

;;; An iconic-text dimensions are not constant.

(defmethod ICON-WIDTH ((icon iconic-text))
  (box-width (icon-box icon)))

(defmethod ICON-HEIGHT ((icon iconic-text))
  (box-height (icon-box icon)))

;;; Draw only the background and text, adjust box when needed.
(defmethod ICON-DRAW ((icon iconic-text))
  (when (should-adjust-p icon)
    (adjust-icon-box icon))
  (icon-draw-background icon)
  (icon-draw-text icon)
  (when (icon-temp-string icon) (icon-draw-string icon)))

;;; Draw the text.
(defmethod ICON-DRAW-TEXT ((icon iconic-text))
  (let ((str (icon-text icon))
        (doc (icon-doc icon))
        (box (icon-box icon)))
    (with-font (doc (text-font icon))
      (with-clipping (doc box)
        (with-foreground-color (doc (text-color icon))
          (with-background-mode (doc :transparent)
            (draw-string-in-box
             doc str 0 (length str) box :center :center nil)))))))

;;; A simple description for text icons.
(defmethod ICON-DESCRIBE ((icon iconic-text) &optional socket)
  (format nil "Text message: ~s" (icon-text icon)))

;;; Do noting on an iconic-text icon-eval.
(defmethod ICON-EVAL ((icon iconic-text))
  (say "Nothing to evaluate.")
  nil)

;;; Add the text color and font of text icons to the icon-style.

(defmethod ICON-STYLE ((icon iconic-text))
  (list* :text-font  (text-font  icon)
         :text-color (text-color icon)
         (call-next-method)))

(defmethod (setf ICON-STYLE) (style (icon iconic-text))
  (when (getf style :text-font)
    (setf (text-font  icon) (getf style :text-font)))
  (when (getf style :text-color)
    (setf (text-color icon) (getf style :text-color)))
  (call-next-method))

;;; Methods to have an iconic-text behave like a normal icon.

(defmethod ICON-LINES-SHOW  ((icon iconic-text))                           t)

(defmethod ICON-LINES-HIDE  ((icon iconic-text) &optional (use-old-box t)) t)

(defmethod ICON-FIND-SOCKET ((icon iconic-text) (pos position))            nil)

(defmethod DELETE-LINKS     ((icon iconic-text))                           t)


;;;-------------------------------------------------------------------------;;;
;;;  I C O N I C - T E X T  icon menu                                       ;;;
;;;-------------------------------------------------------------------------;;;

;;; The menu variable.
(defvar *TEXT-ICON-MENU* nil)

;;; Return this menu for text icons.
(defmethod ICON-MENU ((icon iconic-text))
  *text-icon-menu*)

;;; Create the menu.
(defun MAKE-TEXT-ICON-MENU ()
  (let ((menu (make-default-icon-menu)))
    (add-icon-menu-item menu 'icon-edit-text "Edit ~Text"
                        "Edit the text for this message.")
    menu))

;;; Menu handlers.

;;; Edit the text message.
(defmethod ICON-EDIT-TEXT ((icon iconic-text))
  (let ((newtext (dialog-get-string "Text string" (icon-text icon) t)))
    (when (and newtext
               (not (equalp newtext (icon-text icon)))
               (not (equalp newtext "")))
      (setf (icon-text icon) newtext)
      (do-redraws (icon-doc icon)))))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N I C - T E X T  edit page                                       ;;;
;;;-------------------------------------------------------------------------;;;

;;; The icon-text widgets.
(defvar *ICON-EDIT-ICON-TEXT*        nil)
(defvar *ICON-EDIT-ICON-FONT*        nil)
(defvar *ICON-EDIT-TEXT-COLOR-BOX*   nil)
(defvar *ICON-EDIT-TEXT-COLOR-RED*   nil)
(defvar *ICON-EDIT-TEXT-COLOR-GREEN* nil)
(defvar *ICON-EDIT-TEXT-COLOR-BLUE*  nil)

;;; The iconic-text edit page.
(defun MAKE-TEXT-TAB ()
  (add-icon-edit-tab :text)
  (progrev
   ;; The text box.
   (add-icon-edit-tab-widget
     :widget       'static-text
     :value        "~Text"
     :box          (make-box 10 27 55 52)
     :tabstop      nil
     :groupstart   t)
   (setf *icon-edit-icon-text*
    (add-icon-edit-tab-widget
     :widget       'editable-text
     :value        nil
     :box          (make-box 55 27 230 52)
     :3d-border    t
     :tabstop      t
     :groupstart   nil
     :set-value-fn #'true))
   ;; The font setting.
   (add-icon-edit-tab-widget
     :widget       'static-text
     :value        "~Font"
     :box          (make-box 10 60 55 85)
     :tabstop      nil
     :groupstart   t)
   (setf *icon-edit-icon-font*
    (add-icon-edit-tab-widget
     :widget       'button
     :title        "Font Sample"
     :box          (make-box 55 60 230 85)
     :3d-border    t
     :tabstop      t
     :groupstart   nil
     :set-value-fn #'(lambda (&rest r)
                       (let ((newfont
                              (ask-user-for-font
                               :initial-font (dialog-item-font
                                              *icon-edit-icon-font*))))
                         (when newfont
                           (set-dialog-item-font *icon-edit-icon-font* newfont)
                           nil)))))
   ;; The text color widgets.
   (add-icon-edit-tab-widget
     :widget          'static-text
     :value           "~Color"
     :box             (make-box 10 93 55 118)
     :tabstop         nil
     :groupstart      t)
   (setf *icon-edit-text-color-box*
    (add-icon-edit-tab-widget
     :widget         'color-box
     :value          nil
     :box            (make-box 55 93 82 118)
     :tabstop        nil
     :groupstart     nil
     :set-value-fn   #'color-box-text-set))
   (setf *icon-edit-text-color-red*
    (add-icon-edit-tab-widget
     :widget          'lisp-text
     :name            :red-text-color-text
     :value           nil
     :box             (make-box 85 93 120 118)
     :tabstop         t
     :groupstart      nil
     :up-down-control :red-text-color-up-down
     :set-value-fn    #'color-text-red-set))
   (add-icon-edit-tab-widget
     :widget          'up-down-control
     :name            :red-text-color-up-down
     :value           nil
     :box             (make-box 120 93 130 118)
     :tabstop         nil
     :groupstart      nil
     :buddy-widget    :red-text-color-text
     :buddy-side      :right
     :delayed         nil
     :orientation     :vertical
     :wrapping        nil
     :use-arrow-keys  t
     :increment       nil
     :range           (list 0 255))
   (setf *icon-edit-text-color-green*
    (add-icon-edit-tab-widget
     :widget          'lisp-text
     :name            :green-text-color-text
     :value           nil
     :box             (make-box 135 93 170 118)
     :tabstop         t
     :groupstart      nil
     :up-down-control :green-text-color-up-down
     :set-value-fn    #'color-text-green-set))
   (add-icon-edit-tab-widget
     :widget          'up-down-control
     :name            :green-text-color-up-down
     :value           nil
     :box             (make-box 170 93 180 118)
     :tabstop         nil
     :groupstart      nil
     :buddy-widget    :green-text-color-text
     :buddy-side      :right
     :delayed         nil
     :orientation     :vertical
     :wrapping        nil
     :use-arrow-keys  t
     :increment       nil
     :range           (list 0 255))
   (setf *icon-edit-text-color-blue*
    (add-icon-edit-tab-widget
     :widget          'lisp-text
     :name            :blue-text-color-text
     :value           nil
     :box             (make-box 185 93 220 118)
     :tabstop         t
     :groupstart      nil
     :up-down-control :blue-text-color-up-down
     :set-value-fn    #'color-text-blue-set))
   (add-icon-edit-tab-widget
     :widget          'up-down-control
     :name            :blue-text-color-up-down
     :value           nil
     :box             (make-box 220 93 230 118)
     :tabstop         nil
     :groupstart      nil
     :buddy-widget    :blue-text-color-text
     :buddy-side      :right
     :delayed         nil
     :orientation     :vertical
     :wrapping        nil
     :use-arrow-keys  t
     :increment       nil
     :range           (list 0 255))
   ;; The apply button.
   (add-icon-edit-tab-widget
     :widget       'button
     :title        "~A p p l y"
     :box          (make-box 20 125 220 145)
     :3d-border    t
     :tabstop      t
     :groupstart   nil
     :set-value-fn #'(lambda (&rest r)
                       (let ((newval
                              (dialog-item-value *icon-edit-icon-text*))
                             (newfont
                              (dialog-item-font *icon-edit-icon-font*))
                             (newcolor
                              (dialog-item-value *icon-edit-text-color-box*)))
                         (setf (icon-text  *icon-edit-icon*) newval
                               (text-font  *icon-edit-icon*) newfont
                               (text-color *icon-edit-icon*) newcolor)
                         (do-redraws (icon-doc *icon-edit-icon*))
                         nil)))))

;;; The color item handlers.

(defun COLOR-BOX-TEXT-SET (item new old)
  (unless *called-programmatically*
    (let ((*called-programmatically* t))
      (set-dialog-item-value *icon-edit-text-color-red*   (rgb-red   new))
      (set-dialog-item-value *icon-edit-text-color-green* (rgb-green new))
      (set-dialog-item-value *icon-edit-text-color-blue*  (rgb-blue  new))))
  t)

(defun COLOR-TEXT-RED-SET (item new old)
  (cond (*called-programmatically* t)
        ((and (integerp new) (<= 0 new 255))
         (color-update *icon-edit-text-color-box*
                       new
                       (dialog-item-value *icon-edit-text-color-green*)
                       (dialog-item-value *icon-edit-text-color-blue*))
         t)
        (t nil)))

(defun COLOR-TEXT-GREEN-SET (item new old)
  (cond (*called-programmatically* t)
        ((and (integerp new) (<= 0 new 255))
         (color-update *icon-edit-text-color-box*
                       (dialog-item-value *icon-edit-text-color-red*)
                       new
                       (dialog-item-value *icon-edit-text-color-blue*))
         t)
        (t nil)))

(defun COLOR-TEXT-BLUE-SET (item new old)
  (cond (*called-programmatically* t)
        ((and (integerp new) (<= 0 new 255))
         (color-update *icon-edit-text-color-box*
                       (dialog-item-value *icon-edit-text-color-red*)
                       (dialog-item-value *icon-edit-text-color-green*)
                       new)
         t)
        (t nil)))

;;; Initialize the page values.
(defmethod ICON-EDIT-SETUP-PAGE ((icon iconic-text) (id (eql :text))
                                 &optional socket)
  (set-dialog-item-value *icon-edit-icon-text*      (icon-text icon))
  (set-dialog-item-font  *icon-edit-icon-font*      (text-font icon))
  (set-dialog-item-value *icon-edit-text-color-box* (text-color icon)))

;;; Terminate the page.
(defmethod ICON-EDIT-DONE-PAGE ((icon iconic-text) (id (eql :text)) result)
  (when result
    (setf (icon-text  icon) (dialog-item-value *icon-edit-icon-text*)
          (text-font  icon) (dialog-item-font  *icon-edit-icon-font*)
          (text-color icon) (dialog-item-value *icon-edit-text-color-box*))
    (do-redraws (icon-doc icon))))

;;; Return this page for iconic-text.
(defmethod ICON-TAB-IDS ((x iconic-text))
  (list* :text (call-next-method)))


;;;-------------------------------------------------------------------------;;;
;;;  S H A R E D - C L A S S E S  adding                                    ;;;
;;;-------------------------------------------------------------------------;;;

;;; The shared classes - this is called from each type's registration function
;;; to create its shared nodes.
(defun ADD-SHARED-CLASSES ()
  (add-node-class (iconic-identity-class-name *registering-type*) :shared)
  (add-node-class 'iconic-leaf-value                              :shared)
  (add-node-class 'iconic-text                                    :shared))


;;;-------------------------------------------------------------------------;;;
;;;  I N I T I A L I Z A T I O N  function                                  ;;;
;;;-------------------------------------------------------------------------;;;

;;; This actually only initializes the iconic-text menu and edit page.
(defun INITIALIZE-SHARED-CLASSES ()
  (make-text-tab)
  (setf *text-icon-menu* (make-text-icon-menu)))

;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :SHARED-I)

