/* -----------------------------*- Mode: C -*----------------------------- *\
| File:         midi-lib.c                                                  |
| Description:  BOOMS MIDI system file                                      |
| Author:       Eli Barzilay - eli@cs.bgu.ac.il                             |
| Date:         11-8-95                                                     |
\*-------------------------------------------------------------------------*/

/*

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

*/


#include "midi-lib.h"


/*****************************************************************************/
/*** Global data                                                           ***/
/*****************************************************************************/

HMIDIOUT mapper       = 0;     // This will hold the mapper handle
BOOL     midi_is_open = FALSE; // Is the MIDI system initialized?
Event    events[BUFFER_SIZE];  // The events buffer
int      upper_used   = 0;     // The upmost cell used
int      next_cell    = NIL;   // The free cells list head
int      queue        = NIL;   // The event queue head
int      last_in      = NIL;   // The last event inserted ptr
BOOL     processing   = FALSE; // The buffer is processed, don't handle timer
BOOL     buffer_full  = FALSE; // The buffer is full, we're discarding events
UINT     timer        = 0;     // The timer ID


/*****************************************************************************/
/*** General DLL functions                                                 ***/
/*****************************************************************************/

BOOL APIENTRY DllMain(HANDLE hModule, DWORD reason, LPVOID lpReserved)
{
  static int processes = 0;
  switch(reason) {
  case DLL_PROCESS_ATTACH:
    if (!(processes++)) {
      // Initialize DLL
      midi_open();
    }
    break;
  case DLL_PROCESS_DETACH:
    if (!(--processes)) {
      // Finalize DLL
      if (midi_is_open)
        midi_close();
    }
    break;
  }
  return TRUE;
}


/*****************************************************************************/
/*** General WAV file player                                               ***/
/*****************************************************************************/

// Play a WAV file
BOOL DllExport play_sound (LPCSTR str)
{
  return sndPlaySound (str, SND_ASYNC);
}


/*****************************************************************************/
/*** MIDI openning and closing                                             ***/
/*****************************************************************************/

// Is the MIDI system initialized?
BOOL DllExport midi_open_p()
{
  return midi_is_open;
}

// Initialize the MIDI system
BOOL DllExport midi_open()
{
  if (midi_is_open) {
    error ("midi_open: MIDI already opened");
    return FALSE;
  }
  upper_used  = 0;
  next_cell   = NIL;
  queue       = NIL;
  last_in     = NIL;
  processing  = FALSE;
  buffer_full = FALSE;
  timer       = 0;
  if (midiOutOpen(&mapper,(UINT)MIDI_MAPPER,0,0,0) != MMSYSERR_NOERROR) {
    error ("midi_open: Failed midiOutOpen");
    return FALSE;
  } else {
    midi_is_open = TRUE;
    return TRUE;
  }
}

// Shut down the MIDI system
BOOL DllExport midi_close()
{
  UINT result;
  if (!midi_is_open) {
    error ("midi_close: MIDI already closed");
    return FALSE;
  }
  result = midiOutClose (mapper);
  if (result != MMSYSERR_NOERROR)
    error ("midi_close: Failed midiOutClose");
  deactivate_timer();
  midi_is_open = FALSE;
  return (result == MMSYSERR_NOERROR);
}

// Reset the MIDI system
BOOL DllExport midi_reset()
{
  if (midi_is_open)
    midi_close();
  return midi_open();
}


/*****************************************************************************/
/*** MIDI volume setting                                                   ***/
/*****************************************************************************/

BOOL DllExport set_volume (WORD volume)
{
  union {
    WORD  word[2];
    DWORD vol;
  } m;
  if (!midi_is_open) {
    error ("set_volume: MIDI is closed");
    return FALSE;
  }
  m.word[0] = m.word[1] = volume;
  return (midiOutSetVolume(mapper,m.vol) == MMSYSERR_NOERROR);
}

WORD DllExport get_volume()
{
  union {
    WORD  word[2];
    DWORD vol;
  } m;
  if (!midi_is_open) {
    error ("get_volume: MIDI is closed");
    return 0;
  }
  if (midiOutGetVolume(mapper,&(m.vol)) != MMSYSERR_NOERROR) {
    error ("get_volume: Failed midiOutGetVolume");
    return 0;
  }
  return (m.word[0] + m.word[1]) / 2;
}


/*****************************************************************************/
/*** Play a MIDI note                                                      ***/
/*****************************************************************************/

// Using the buffer
void DllExport play_note (DWORD start, UINT duration,
                          char pitch, char velocity)
{
  union {
    BYTE  byte[4];
    DWORD msg;
  } m;
  if (!midi_is_open) {
    error ("play_note: MIDI is closed");
    return;
  }
  m.byte[3] = 0;
  m.byte[0] = 0x90; m.byte[1] = pitch; m.byte[2] = velocity;
  put_event (m.msg, start);
  start += duration;
  m.byte[0] = 0x80; m.byte[1] = pitch; m.byte[2] = velocity;
  put_event (m.msg, start);
}


/*****************************************************************************/
/*** Event queue timing handle                                             ***/
/*****************************************************************************/

// Get the current time
DWORD DllExport get_time()
{
  return timeGetTime();
}

// Put a MIDI event on the queue
void DllExport put_event (DWORD message, DWORD time)
{
  int cell, p, p1;
  if (!midi_is_open) {
    error ("put_event: MIDI is closed");
    return;
  }
  activate_timer();
  cell = get_cell();
  if ((cell == NIL) && (!buffer_full)) {
    buffer_full = TRUE;
    error ("put_event: Event buffer full, events discarded!");
    return;
  }
  buffer_full = FALSE;
  processing = TRUE;
  events[cell].message = message;
  events[cell].time    = time;
  if ((queue == NIL) || (events[queue].time >= time)) {
    events[cell].next = queue;
    queue = cell;
    last_in = NIL;
  } else {
    if ((last_in != NIL) && (time >= events[last_in].time))
      p1 = last_in;
    else
      p1 = queue;
    p = events[p1].next;
    while ((p != NIL) && (events[p].time < time)) {
      p1 = p;
      p = events[p].next;
    }
    events[cell].next = p;
    events[p1].next = cell;
    last_in = cell;
  }
  processing = FALSE;
}


/*****************************************************************************/
/*** Timer handling                                                        ***/
/*****************************************************************************/

// Activate the timer
void activate_timer()
{
  if (timer)
    return;
  if (timeBeginPeriod(RESOLUTION) != TIMERR_NOERROR) {
    error ("activate_timer: Failed timeBeginPeriod");
    return;
  }
  timer = timeSetEvent (RESOLUTION, RESOLUTION,
                        (LPTIMECALLBACK)(&TimerCallBack), 0, TIME_PERIODIC);
  if (timer == (UINT)NULL) {
    error ("activate_timer: Failed timeSetEvent");
    if (timeEndPeriod(RESOLUTION != TIMERR_NOERROR))
      error ("activate_timer: Failed timeEndPeriod");
  }
}

// Deactivate the timer
void deactivate_timer()
{
  if (!timer)
    return;
  if (timeKillEvent(timer) != TIMERR_NOERROR)
    error ("deactivate_timer: Failed timeKillEvent");
  else if (timeEndPeriod(RESOLUTION) != TIMERR_NOERROR)
    error ("deactivate_timer: Failed timeEndPeriod");
  timer = 0;
}


/*****************************************************************************/
/*** Event queue timing handle                                             ***/
/*****************************************************************************/

// The callback function - pop events andsend them to MIDI
void CALLBACK TimerCallBack (UINT id, UINT msg, DWORD usr, DWORD w1, DWORD w2)
{
  DWORD now;
  int   p;
  if (!processing) {
    now = timeGetTime();
    while ((queue != NIL) && (events[queue].time <= now)) {
      midi_send (events[queue].message);
      p = queue;
      if (last_in == queue)
        last_in = NIL;
      queue = events[queue].next;
      free_cell (p);
    }
    if (queue == NIL)
      deactivate_timer;
  }
}


/*****************************************************************************/
/*** Hadling internal event allocation                                     ***/
/*****************************************************************************/

// Allocate a cell on the buffer
int get_cell()
{
  int cell;
  if (next_cell != NIL) {
    cell = next_cell;
    next_cell = events[next_cell].next;
  } else if (upper_used < BUFFER_SIZE)
    cell = upper_used++;
  else
    cell = NIL;
  return cell;
}

// Free a buffer cell
void free_cell (int cell)
{
  events[cell].next = next_cell;
  next_cell = cell;
}


/*****************************************************************************/
/*** Sending MIDI messages                                                 ***/
/*****************************************************************************/

// Send one MIDI message
BOOL DllExport midi_send (DWORD Msg)
{
  if (!midi_is_open) {
    error ("midi_send: MIDI is closed");
    return FALSE;
  }
  return (midiOutShortMsg(mapper, Msg) == MMSYSERR_NOERROR) ? TRUE : FALSE;
}


/*****************************************************************************/
/*** Error message box                                                     ***/
/*****************************************************************************/

// Display an error box
void error (LPCSTR str)
{
  MessageBox (NULL, str, "MIDI-LIB Message",
              MB_APPLMODAL | MB_ICONEXCLAMATION | MB_OK);
}


/*****************************************************************************/

